# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module contains classes related with EM
"""

import os
from pkg_resources import parse_version

import pyworkflow as pw
from pyworkflow.protocol import Protocol
from pyworkflow.utils import  getSubclasses
from pyworkflow.viewer import Viewer
from pyworkflow.wizard import Wizard
import pyworkflow.plugin

from .constants import *
from .objects import EMObject
from .tests import defineDatasets
from .utils import *

__version__ = '3.3.1'
NO_VERSION_FOUND_STR = "0.0"
CUDA_LIB_VAR = 'CUDA_LIB'

_logo = "scipion_icon.png"
_references = ["delaRosaTrevin201693"]


class Config(pw.Config):
    _get = pw.Config._get
    _join = pw.Config._join

    EM_ROOT = _join(_get(EM_ROOT_VAR, _join(pw.Config.SCIPION_SOFTWARE, 'em')))

    # Default XMIPP_HOME: needed here for ShowJ viewers
    XMIPP_HOME = _join(_get('XMIPP_HOME', os.path.join(EM_ROOT, 'xmipp')))

    # Get java home, we might need to provide correct default value. Use SCIPION_JAVA_HOME to force it when there is other JAVA_HOME you don't want/cant to change: e.g. pycharm debugging.
    JAVA_HOME = _get('SCIPION_JAVA_HOME', _get('JAVA_HOME', ''))
    JAVA_MAX_MEMORY = _get('JAVA_MAX_MEMORY', '4')

    # MPI
    MPI_LIBDIR = _get('MPI_LIBDIR', '/usr/lib64/mpi/gcc/openmpi/lib')
    MPI_BINDIR = _get('MPI_BINDIR', '/usr/lib64/mpi/gcc/openmpi/bin')

    # CUDA
    CUDA_LIB = _get(CUDA_LIB_VAR, '/usr/local/cuda/lib64')
    CUDA_BIN = _get('CUDA_BIN', '/usr/local/cuda/bin')
    MAX_PREVIEW_FILE_SIZE = float(_get("MAX_PREVIEW_FILE_SIZE", DEFAULT_MAX_PREVIEW_FILE_SIZE))

    # OLD CHIMERA variable
    CHIMERA_OLD_BINARY_PATH = _get("CHIMERA_OLD_BINARY_PATH",'')

    # Path to either ImageJ or Fiji binary program
    IMAGEJ_BINARY_PATH = _get("IMAGEJ_BINARY_PATH",'')


class Domain(pyworkflow.plugin.Domain):
    _name = __name__
    _objectClass = EMObject
    _protocolClass = Protocol
    _viewerClass = Viewer
    _wizardClass = Wizard
    _baseClasses = getSubclasses(EMObject, globals())


class Plugin(pyworkflow.plugin.Plugin):
    _url = URL

    @classmethod
    def _defineEmVar(cls, varName, defaultValue):
        """ Shortcut method to define variables prepending EM_ROOT if variable is not absolute"""

        # Get the value, either whatever is in the environment or a join of EM_ROOT + defaultValue
        value = os.environ.get(varName, os.path.join(Config.EM_ROOT, defaultValue))

        # CASE-1 : Users might have used ~ and that has to be expanded
        value = os.path.expanduser(value)

        # CASE-2 :Old configs 2.0 will likely have:
        # EM_ROOT = software/em
        # CHIMERA_HOME = %(EM_ROOT)s/chimera-13.0.1
        #  ...
        #
        # this end up in the environment resolved like: software/em/chimera-13.0.1
        # whereas as default values will come as absolute:
        # /<scipion_home>/software/em/chimera-13.0.1
        # In any case we join it (absolute paths will not join)
        value = os.path.join(pw.Config.SCIPION_HOME, value)

        cls._addVar(varName, value)

    @classmethod
    def getMaxitHome(cls):
        return cls.getVar(MAXIT_HOME)

    @classmethod
    def getMaxitBin(cls):
        return os.path.join(cls.getMaxitHome(), 'bin', MAXIT)

    @classmethod
    def _defineVariables(cls):
        # Avoid defining variables from children that does not define variables.
        if cls == Plugin:
            cls._defineVar(EM_ROOT_VAR, pwem.Config.EM_ROOT)
            cls._defineEmVar(MAXIT_HOME, 'maxit-10.1')

            # Take this initialization event to define own datasets
            defineDatasets()
            # Register filehandlers too
            cls._registerFileHandlers()

    @classmethod
    def defineBinaries(cls, env):
        cls.defineBinariesMaxit(False, env)

    @classmethod
    def defineBinariesMaxit(cls, default, env):
        # If not defined already (several plugins needs this and call this but has to be added once
        if not env.hasPackage(MAXIT):
            MAXIT_URL = 'https://sw-tools.rcsb.org/apps/MAXIT/maxit-v10.100-prod-src.tar.gz'
            MAXIT_TAR = 'maxit-v10.100-prod-src.tar.gz'
            maxit_commands = [('make -j 1 binary ', ['bin/maxit'])]
            env.addPackage(MAXIT, version='10.1',
                           tar=MAXIT_TAR,
                           url=MAXIT_URL,
                           neededProgs=['gcc', 'flex', 'make', 'bison', 'tcsh'],
                           commands=maxit_commands,
                           default=default)  # scipion installb maxit
            # requirements bison, flex, gcc

        maxit10 = env.getTarget(env._getExtName(MAXIT, "10.1"))
        maxit10.setDefault(default)


    @classmethod
    def guessCudaVersion(cls, variable, default="10.1"):
        """ Guesses the cuda version from a variable

        :param variable: Name of a variable defined by a plugin
        :param default: Default value in case cuda is not found
        :return Version after parsing the variable or default"""

        return cls.getVersionFromVariable(variable, default=default, pattern="cuda")

    @classmethod
    def getVersionFromVariable(cls, variable, default=NO_VERSION_FOUND_STR, pattern=None):
        """ Returns a Version instance from parsing the content of
         a variable

         :param variable: Variable name to use and parse
         :param default: Optional. Default value to return in case version is not found
         :param pattern: Optional. If passed, the pattern will be used to locate the folder
         :return Version after parsing the variable"""

        value = cls.getVar(variable)
        return cls.getVersionFromPath(value, default= default, pattern=pattern)


    @classmethod
    def getVersionFromPath(cls, path, separator="-", default=NO_VERSION_FOUND_STR, pattern=None):
        """ Resolves path to the realpath (links) and returns the last part
         of the path after the separator as a Version object.
         If separator is not present returns None

         :param path: string containing a path with a version as part of its name
         :param separator: Optional. defaults to "-".
         :param default: Optional. Default version ("X.Y" string) to return if no version is found
         :param pattern: Optional. If passed, the pattern will be used to locate the folder
         :return Version object filled with version found or 0.0

         """

        path = os.path.realpath(path)
        if pattern is not None:
            path = findFolderWithPattern(path, pattern)
            if path is None:
                return parse_version(default)

        parts = path.split(separator)
        if len(parts)>=2:

            # Version should be the last bit
            versionStr = parts[-1]
            return parse_version(versionStr)
        else:
            return parse_version(default)
    @classmethod
    def _registerFileHandlers(cls):
        # register file handlers to preview info in the Filebrowser....
        from pyworkflow.gui.browser import FileTreeProvider, STANDARD_IMAGE_EXTENSIONS
        from .viewers.filehandlers import MdFileHandler, ParticleFileHandler, VolFileHandler, StackHandler, ChimeraHandler, ImajeJFileHandler

        register = FileTreeProvider.registerFileHandler
        register(MdFileHandler(), '.xmd', '.star', '.pos', '.ctfparam', '.doc')
        register(ParticleFileHandler(),
                 '.xmp', '.tif', '.tiff', '.spi', '.mrc', '.map', '.raw',
                 '.inf', '.dm3', '.em', '.pif', '.psd', '.spe', '.ser', '.img',
                 '.hed', *STANDARD_IMAGE_EXTENSIONS)
        register(VolFileHandler(), '.vol', '.hdf', '.rec')
        register(StackHandler(), '.stk', '.mrcs', '.st', '.pif', '.dm4', '.ali')
        register(ChimeraHandler(), '.bild', '.mrc', '.pdb', '.vol', '.hdf', '.cif', '.mmcif')

        if Config.IMAGEJ_BINARY_PATH:
            register(ImajeJFileHandler(), '.mrcs', '.mrc', '.st', '.ali', '.rec', '.tif', '.tiff', *STANDARD_IMAGE_EXTENSIONS)
        else:
            msg= "Optional: ImageJ of Fiji not configured can be configured to open files from the File Browser." \
                 " Please add 'IMAGEJ_BINARY_PATH' to the config file (%s)" \
                 "and point to the binary file that will open the files." % pw.Config.SCIPION_CONFIG
            logger.info(msg)



def findFolderWithPattern(path, pattern):
    """
    Returns the first folder found from the right that matches the pattern

    :param path: path to do the search on
    :param pattern: pattern to look for. No regex for now, just a contains

    :return the folder matching the pattern or None

    """

    previous, last = os.path.split(path)

    # If end reached: windows Paths and linux paths seems to end in a different way
    if previous == "" or last=="":
        return  None
    elif pattern in last:
        return last
    else:
        return findFolderWithPattern(previous, pattern)




# **************************************************************************
# *
# * Authors:     Pablo Conesa [1]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.utils as pwutils
import pyworkflow.protocol.params as params

import pwem.objects as emobj
from pwem.protocols import EMProtocol
import logging
logger = logging.getLogger(__file__)

OUTPUT_NAME = "output"


class ProtClassesSelector(EMProtocol):
    """ Extracts items from a SetOfClasses based on number of items assigned to the class
    """

    _label = "numeric classes extractor"
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('inputClasses', params.PointerParam,
                      pointerClass='SetOfClasses',
                      label='Input classes',
                      help='Set of classes to extract items from.')
        form.addParam('extractRepresentative', params.BooleanParam,
                      label='Extract representative?',
                      help='Set to true if you want to extract the image that represents the class otherwise all '
                           'items supporting the class will be extracted.',
                      default = False)
        form.addParam('firstNElements', params.IntParam,
                      label='Biggest N classes',
                      help='Will take the N biggest classes. Those having most element supporting them.',
                      default=5)


# -------------------------- INSERT steps functions ---------------------------

    def _insertAllSteps(self):
        """ Insert all steps
        """
        self._insertFunctionStep(self.extractElements.__name__)

    def extractElements(self):

        count= 0
        # For each class (order by number of items)
        for clazz in self.inputClasses.get().iterItems(orderBy="_size", direction="DESC"):

            # Increment the count
            count += 1

            # If haven't reached the Nelements
            if count <= self.firstNElements.get():

                print(clazz.getSize())
                self._extractElementFromClass(clazz)

        output = self._getOutputSet()
        output.write()
        self._store(output)

    def _extractElementFromClass(self, clazz):

        output = self._getOutputSet()

        # If taking representatives
        if self.extractRepresentative.get():
            rep = clazz.getRepresentative().clone()
            output.append(rep)
        else:
            # Go through all items and append them
            for image in clazz:
                newImage = image.clone()
                output.append(newImage)

    def _getOutputSet(self):
        """ Returns the output set creating it if not yet done"""

        # If output not created yet
        if not hasattr(self, OUTPUT_NAME):
            outputSet = None

            # Create a set ot items or a set of representatives
            if self.extractRepresentative.get():

                self.info("Creating set from representatives.")
                outputSet = createSetFromRepresentative(self.inputClasses.get(), self._getPath())
            else:
                self.info("Creating set from images.")
                outputSet = createSetFromImages(self.inputClasses.get(), self._getPath())

            self._defineOutputs(**{OUTPUT_NAME:outputSet})

        return getattr(self, OUTPUT_NAME)


    def _summary(self):
        summary = []
        summary.append("%s option selected." % "representative/s" if self.extractRepresentative else "items")
        summary.append("%s biggest classes to be extracted from." % self.firstNElements.get())
        return summary

    def _methods(self):
        extractedElement = "representative/s" if self.extractRepresentative else "items"
        methods = ["The %s of the %s biggest classes were selected from %s." % (extractedElement, self.firstNElements.get(), self.getObjectTag(self.inputClasses))]
        return methods

    def _validate(self):
        errors = []
        return errors

# Helpers
def createSetFromRepresentative(classesSet, path):
    """ Creates a teh corresponding set from the representative of a set of classes"""

    rep_type = classesSet.REP_TYPE

    # need to instantiate the right set based on the representative.
    # Going from the REP_TYPE to the right set is not straight forward. e.g:
    # SetOfClasses2D --> REP_TYPE = Particle --> Set for the rep type == SetOfAverages
    # So we need to make do some manual checks. TODO: have factory methods in the SetOfClasses for the representatives?
    setClass = None
    logger.debug("Creating representative set from %s: type %s" % (classesSet.__class__, rep_type))
    if rep_type.__name__ ==  emobj.Particle.__name__:
        setClass = emobj.SetOfAverages
    else:
        setClass = pwutils.Config.getDomain().getObjects()["SetOf%ss" % rep_type.__name__]
    set = setClass.create(outputPath=path)

    imgs = classesSet.getImages()
    set.copyInfo(imgs)

    return set

def createSetFromImages(classesSet, path):
    """ Creates a the corresponding set from the images of a set of classes"""

    images = classesSet.getImages()

    # need to instantiate the right set based on the Items.

    setClass = None
    logger.debug("Creating an image set from %s: type %s" % (classesSet.__class__, images))

    setClass = pwutils.Config.getDomain().getObjects()[images.__class__.__name__]

    set = setClass.create(outputPath=path)

    set.copyInfo(images)

    return set






import unittest
from unittest.mock import patch

from pwem import Plugin, findFolderWithPattern, CUDA_LIB_VAR, NO_VERSION_FOUND_STR


class TestPluginCudaGuess(unittest.TestCase):
    """ Tests code that tries to guess what is the CUDA version and functionality related"""

    def test_cuda_internal_methods(self):

        # findFolderWithPattern
        self.assertIsNone(findFolderWithPattern("/one/two/three", "four"))
        self.assertEqual(findFolderWithPattern("/one/two/three", "tw"), "two")

        v = Plugin.getVersionFromPath("/usr/local/cuda-10.2")
        self.assertEqual(v.major, 10, "Major not parsed from path with version")
        self.assertEqual(v.minor, 2, "Minor not parsed from path with version")

        v = Plugin.getVersionFromPath("/usr/local/cuda10.2")
        self.assertEqual(str(v), NO_VERSION_FOUND_STR, "Default value not working")

    def test_cuda_version_in_variable(self):

        with patch("os.path.realpath") as realpath_faked:
            realpath_faked.return_value = "/usr/local/cuda-13.8/lib-12.4"

            with patch.object(Plugin, "getVar") as getVarFaked:

                # Mock getVar return value
                getVarFaked.return_value = "/usr/local/useless-11.6"

                v = Plugin.getVersionFromVariable(CUDA_LIB_VAR)
                self.assertEqual(v.major, 12, "Major not parsed from link path")
                self.assertEqual(v.minor, 4, "Minor not parsed from link path")

                v = Plugin.getVersionFromVariable(CUDA_LIB_VAR, pattern="cuda")
                self.assertEqual(v.major, 13, "Major not parsed from link path with a pattern")
                self.assertEqual(v.minor, 8, "Minor not parsed from link path with a pattern")

                v = Plugin.guessCudaVersion(CUDA_LIB_VAR)
                self.assertEqual(v.major, 13, "Major not parsed from link path with a pattern")
                self.assertEqual(v.minor, 8, "Minor not parsed from link path with a pattern")

                # Mock getVar return value. Grigory's case ;-)
                realpath_faked.return_value = "/usr/lib/x86_64-linux-gnu"

                v = Plugin.guessCudaVersion(CUDA_LIB_VAR)
                self.assertEqual(str(v), "10.1", "Guess cuda version does not return default=10.1")



# **************************************************************************
# *
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.protocol.params as params

from .protocol_2d import ProtAlign2D
from pyworkflow.utils import ProgressBar


class ProtAlignmentInvertHand(ProtAlign2D):
    """ Modify the transformation matrix of a set of particles
    So that the handedness changes
    """
    _label = 'invert hand'

    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('inputParticles', params.PointerParam, pointerClass='SetOfParticles',
                      label='Input particles',
                      help='Select the particles that you want to update the new alignment.')
        form.addParallelSection(threads=0, mpi=0)

# --------------------------- INSERT steps functions --------------------------

    def _insertAllSteps(self):
        """for each ctf insert the steps to compare it
        """
        self._insertFunctionStep('createOutputStep')

    def createOutputStep(self):
        inputParticles = self.inputParticles.get()
        outputParticles = self._createSetOfParticles()
        outputParticles.copyInfo(inputParticles)
        outputParticles.setAlignment(inputParticles.getAlignment())
        total = len(inputParticles)
        width = 40
        step = total // width

        # print a progressbar since this make take for ever
        pb = ProgressBar(total=total, width=width, fmt=ProgressBar.NOBAR)
        for counter, particle in enumerate(inputParticles):
            if counter % step == 0:
                pb.update(counter)
            t = particle.getTransform()
            m = t.getMatrix()
            m[0, 2] *= -1.
            m[1, 2] *= -1.
            m[2, 1] *= -1.
            m[2, 0] *= -1.
            # particle.setTransform(t.setMatrix(m))
            outputParticles.append(particle)
        pb.finish()

        self._defineOutputs(outputParticles=outputParticles)
        self._defineSourceRelation(self.inputParticles, outputParticles)

    def _summary(self):
        summary = []
        if not hasattr(self, 'outputParticles'):
            summary.append("Output particles not ready yet.")
        return summary

    def _methods(self):
        methods = []
        if not hasattr(self, 'outputParticles'):
            methods.append("Output particles not ready yet.")
        return methods

    def _validate(self):
        """ The function of this hook is to add some validation before the protocol
        is launched to be executed. It should return a list of errors. If the list is
        empty the protocol can be executed.
        """
        # check that input set of aligned particles do have 2D alignment
        errors = []
        inputAlignmentSet = self.inputParticles.get()
        if not inputAlignmentSet.hasAlignment():
            errors.append("Input alignment set should contains some kind of alignment (2D, 3D or Projection).")
        else:
            # Just for consistency, check that the particles really contains Transform object
            first = inputAlignmentSet.getFirstItem()
            alignment = first.getTransform()
            if alignment is None:
                errors.append('Inconsistency detected in *Input alignment* !!!')
                errors.append('It has alignment: _%s_, but the alignment is missing!!!' %
                              inputAlignmentSet.getAlignment())
            
        # Add some errors if input is not valid
        return errors



# **************************************************************************
# *
# * Authors:     Pablo Conesa(pconesa@cnb.csic.es)
# *              Roberto Marabini(roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.protocol.params as params
from pwem.protocols import EMProtocol
from pwem.objects.data import SetOfParticles

import numpy as np


class ProtSetEditor(EMProtocol):
    """
    Protocol to edit attributes of all the items of a set using a formula.
    This could be useful for editing some values in the set. Use this
    protocol with extreme care, you can easily produce a set that is
    not consistent.
    """
    _label = 'edit set'

    def _defineParams(self, form):
        """
        Defines the parameters the protocol form will show and its behaviour
        :param form:
        """
        form.addSection(label='Input')
        form.addParam('inputSet', params.PointerParam,
                      pointerClass='EMSet',
                      label='Set to edit',
                      help='Set which items will be modified.')
        # formula
        form.addParam('formula', params.StringParam, label="Formula",
                      help='A python code compatible with eval, where item represents each of '
                           'the elements of the set. E.g.: item._resolution.set(item._resolution.get() +1).'
                           'You could also use modules like "import numpy;  item._resolution .... "')

    def _insertAllSteps(self):
        self._insertFunctionStep('formulaStep')

    def formulaStep(self):
        """
        Goes through all items in the input set and applies the formula to each of them using exec.
        Complex python code could be run separating lines with ;  To use numpy you could do
        import numpy; item._resolution.set(numpy.random.randint(10))
        """
        inputSet = self.inputSet.get()
        modifiedSet = inputSet.createCopy(self._getExtraPath(), copyInfo=True)

        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            exec(self.formula.get())
            modifiedSet.append(item)

        self.createOutput(self.inputSet, modifiedSet)

    def createOutput(self, inputSet, modifiedSet):
        """ Save the output set."""
        outputArgs = {inputSet.getExtended(): modifiedSet}
        self._defineOutputs(**outputArgs)

    def _summary(self):

        return ["The this formula (%s) is/was applied to the items of the input set." % self.formula.get()]





# **************************************************************************
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import pyworkflow.tests as pwtests

import pwem.protocols as emprot

from .test_workflow import TestWorkflow


# update this test when RCT workflow are implemented
class TestXmippRCTWorkflow(TestWorkflow):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dataset = pwtests.DataSet.getDataSet('rct')
        cls.allCrdsDir = cls.dataset.getFile('positions')
        cls.micsUFn = cls.dataset.getFile('untilted')
        cls.micsTFn = cls.dataset.getFile('tilted')
        cls.classesSqlite = cls.dataset.getFile('classes')

    def testXmippRCTWorkflowBasic(self):
        # First, import a set of micrographs
        protImport = self.newProtocol(emprot.ProtImportMicrographsTiltPairs,
                                      patternUntilted=self.micsUFn,
                                      patternTilted=self.micsTFn,
                                      samplingRate=2.28, voltage=100,
                                      sphericalAberration=2.9)
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographsTiltPair,
                             "There was a problem with the import")
        # self.validateFiles('protImportRCT', protImport)

        # Then simulate a particle picking
        print("Running fake particle picking...")
        from pwem import Domain
        XmippProtParticlePickingPairs = Domain.importFromPlugin(
            'xmipp3.protocols',
            'XmippProtParticlePickingPairs',
            doRaise=True)
        protPicking = self.newProtocol(XmippProtParticlePickingPairs,
                                       importFolder=self.allCrdsDir)

        protPicking.inputMicrographsTiltedPair.set(protImport.outputMicrographsTiltPair)

        self.proj.launchProtocol(protPicking, wait=True)

        self.assertIsNotNone(protPicking.outputCoordinatesTiltPair,
                             "There was a problem with the faked picking")
        # self.validateFiles('protPicking', protPicking)

        # Extract particles
        print("Run extract particles with Same as picking")
        XmippProtExtractParticlesPairs = Domain.importFromPlugin(
            'xmipp3.protocols',
            'XmippProtExtractParticlesPairs')
        SAME_AS_PICKING = Domain.importFromPlugin('xmipp3.constants',
                                                  'SAME_AS_PICKING')
        protExtract = self.newProtocol(XmippProtExtractParticlesPairs,
                                       downFactor=2,
                                       boxSize=60,
                                       doInvert=False,
                                       downsampleType=SAME_AS_PICKING)
        protExtract.inputCoordinatesTiltedPairs.set(protPicking.outputCoordinatesTiltPair)

        self.proj.launchProtocol(protExtract, wait=True)

        # self.validateFiles('protExtract', protExtract)
        self.assertIsNotNone(protExtract.outputParticlesTiltPair,
                             "There was a problem with the extract particles")

        # Classify using Xmipp CL2D
        print("Run CL2D")
        XmippProtCL2D = Domain.importFromPlugin('xmipp3.protocols',
                                                'XmippProtCL2D')
        protCL2D = self.newProtocol(XmippProtCL2D,
                                    numberOfClasses=10,
                                    numberOfInitialClasses=1,
                                    numberOfIterations=3, numberOfMpi=2)
        protCL2D.inputParticles.set(protExtract.outputParticlesTiltPair.getUntilted())
        self.launchProtocol(protCL2D)
        self.assertIsNotNone(protCL2D.outputClasses,
                             "There was a problem with CL2D")
        # self.validateFiles('protCL2D', protCL2D)

        # Random Conical Tilt
        print("Run Random Conical Tilt")
        XmippProtRCT = Domain.importFromPlugin('xmipp3.protocols',
                                               'XmippProtRCT')
        protRCT = self.newProtocol(XmippProtRCT)
        protRCT.inputParticlesTiltPair.set(protExtract.outputParticlesTiltPair)
        protRCT.inputParticles.set(protCL2D.outputClasses)

        self.proj.launchProtocol(protRCT, wait=True)

        # self.validateFiles('protExtract', protExtract)
        self.assertIsNotNone(protRCT.outputVolumes, "There was a problem with the RCT")



# **************************************************************************
# *
# * Authors:    Amaya Jimenez (ajimenez@cnb.csic.es)
# *              Marta Martinez (mmmtnez@cnb.csic.es)
# *              Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

import pyworkflow.tests as pwtest

import pwem.protocols as emprot
import pwem.constants as emcts
from .test_protocols_import_volumes import createFeatVolume


class TestExport2DataBases(pwtest.BaseTest):
    @classmethod
    def runImportVolumes(cls, pattern, samplingRate, label):
        """ Run an Import volumes protocol. """
        cls.protImport = cls.newProtocol(emprot.ProtImportVolumes,
                                         objLabel=label,
                                         filesPath=pattern,
                                         samplingRate=samplingRate
                                         )
        cls.launchProtocol(cls.protImport)
        return cls.protImport

    @classmethod
    def setData(cls, dataProject='resmap'):
        cls.dataset = pwtest.DataSet.getDataSet(dataProject)
        cls.half1 = cls.dataset.getFile('betagal_half1')
        cls.half2 = cls.dataset.getFile('betagal_half2')
        cls.dsModBuild = pwtest.DataSet.getDataSet('model_building_tutorial')

    @classmethod
    def setUpClass(cls):
        pwtest.setupTestProject(cls)
        cls.setData()
        cls.protImportHalf1 = cls.runImportVolumes(cls.half1, 3.54,
                                                   'import half1')
        cls.protImportHalf2 = cls.runImportVolumes(cls.half2, 3.54,
                                                   'import half2')

    @classmethod
    def _importAtomStructCIF(self):
        args = {'inputPdbData': emprot.ProtImportPdb.IMPORT_FROM_FILES,
                'pdbFile': self.dsModBuild.getFile(
                    'PDBx_mmCIF/5ni1.cif'),
                }
        protImportPDB = self.newProtocol(emprot.ProtImportPdb, **args)
        protImportPDB.setObjLabel('import atom struct\nmmCIF\n5ni1.cif')
        self.launchProtocol(protImportPDB)
        structure_mmCIF = protImportPDB.outputPdb
        return structure_mmCIF

    def test_volume1_AtomStructCIF(self):
        from pwem import Domain
        XmippProtResolution3D = Domain.importFromPlugin('xmipp3.protocols',
                                                        'XmippProtResolution3D',
                                                        doRaise=True)

        prot = self.newProtocol(XmippProtResolution3D)
        prot.inputVolume.set(self.protImportHalf1.outputVolume)
        prot.referenceVolume.set(self.protImportHalf2.outputVolume)
        self.launchProtocol(prot)

        protExp = self.newProtocol(emprot.ProtExportDataBases)
        protExp.setObjLabel("export to DB\ndir1")
        protExp.exportVolume.set(self.protImportHalf1.outputVolume)
        protExp.exportFSC.set(prot.outputFSC)
        protExp.exportAtomStruct.set(self._importAtomStructCIF())

        protExp.filesPath.set(os.getcwd() + "/dir1")
        self.launchProtocol(protExp)

        # Check the files were generated properly.
        # protExp._createFileNamesTemplates()
        dirName = protExp.filesPath.get()
        nameVolume = os.path.join(dirName, protExp.VOLUMENAME)
        nameFsc = os.path.join(dirName, "fsc_%02d.xml" % 1)
        nameAtomStruct = os.path.join(dirName, protExp.COORDINATEFILENAME)
        self.assertTrue(os.path.exists(nameVolume))
        self.assertTrue(os.path.exists(nameFsc))
        self.assertTrue(os.path.exists(nameAtomStruct))

        # Check if the files have the correct data
        orig_list_x, orig_list_y = protExp.exportFSC.get().getData()
        fo = open(nameFsc, "rU")
        saved_x = []
        orig_x = []
        count = 0
        for line in fo:
            if line[0:3] == '<x>':
                saved_x.append(int(float(line[3:-5]) * 1000))
                orig_x.append(int(orig_list_x[count] * 1000))
                count = count + 1

        self.assertListEqual(orig_x, saved_x)

    def test_volume1_AtomStructPDB(self):
        from pwem import Domain
        XmippProtResolution3D = Domain.importFromPlugin('xmipp3.protocols',
                                                        'XmippProtResolution3D',
                                                        doRaise=True)
        prot = self.newProtocol(XmippProtResolution3D)
        prot.inputVolume.set(self.protImportHalf1.outputVolume)
        prot.referenceVolume.set(self.protImportHalf2.outputVolume)
        self.launchProtocol(prot)

        protExp = self.newProtocol(emprot.ProtExportDataBases)
        protExp.setObjLabel("export to DB\ndir2")
        protExp.exportVolume.set(self.protImportHalf1.outputVolume)
        protExp.exportFSC.set(prot.outputFSC)

        # run Chimera rigid fit ti get a PDB file
        extraCommands = ""
        extraCommands += "scipionwrite #3 prefix DONOTSAVESESSION_\n"
        extraCommands += "exit\n"

        args = {'extraCommands': extraCommands,
                'inputVolume': self.protImportHalf1.outputVolume,
                'pdbFileToBeRefined': self._importAtomStructCIF()
                }
        from pwem import Domain

        ChimeraProtRigidFit = Domain.importFromPlugin('chimera.protocols',
                                                      'ChimeraProtRigidFit',
                                                      doRaise=True)
        protChimera = self.newProtocol(ChimeraProtRigidFit, **args)
        protChimera.setObjLabel('chimera fit\n volume and PDB\n save model')
        self.launchProtocol(protChimera)

        outputAttr = "DONOTSAVESESSION_Atom_struct__3_%06d" % protChimera.getObjId()
        protExp.exportAtomStruct.set(getattr(protChimera, outputAttr))

        protExp.filesPath.set(os.getcwd() + "/dir2")
        self.launchProtocol(protExp)

        # Check the files were generated properly.
        # protExp._createFileNamesTemplates()
        dirName = protExp.filesPath.get()
        nameVolume = os.path.join(dirName, protExp.VOLUMENAME)
        nameFsc = os.path.join(dirName, "fsc_%02d.xml" % 1)
        nameAtomStruct = os.path.join(dirName, protExp.COORDINATEFILENAME)
        self.assertTrue(os.path.exists(nameVolume))
        self.assertTrue(os.path.exists(nameFsc))
        self.assertTrue(os.path.exists(nameAtomStruct))

        # Check if the files have the correct data
        orig_list_x, orig_list_y = protExp.exportFSC.get().getData()
        fo = open(nameFsc, "rU")
        saved_x = []
        orig_x = []
        count = 0
        for line in fo:
            if line[0:3] == '<x>':
                saved_x.append(int(float(line[3:-5]) * 1000))
                orig_x.append(int(orig_list_x[count] * 1000))
                count = count + 1

        self.assertListEqual(orig_x, saved_x)

    def import_volume_halfmaps(self):
        """
        Test to import a full map (Icosahedron) and two maps (half1 and half2)
        to compute the FSC
        """
        volFeatName = '/tmp/Icosahedron_map.txt'
        volMapNamefull = '/tmp/Icosahedron_map_full.mrc'
        volMapNamehalf1 = '/tmp/Icosahedron_map_half1.mrc'
        volMapNamehalf2 = '/tmp/Icosahedron_map_half2.mrc'
        volMaskName = '/tmp/mask.mrc'
        createFeatVolume(volFeatName, volMapNamefull, sym=emcts.SYM_I222r)
        createFeatVolume(volFeatName, volMapNamehalf1, sym=emcts.SYM_I222r)
        createFeatVolume(volFeatName, volMapNamehalf2, sym=emcts.SYM_I222r, factor=1.005)
        createFeatVolume(volFeatName, volMaskName, sym=emcts.SYM_I222r, factor=1.2)
        _samplingRate = 1.0

        # import 3d map
        args = {'filesPath': volMapNamefull,
                'filesPattern': '',
                'setHalfMaps': True,
                'half1map': volMapNamehalf1,
                'half2map': volMapNamehalf2,
                'samplingRate': _samplingRate
                }
        prot5 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot5.setObjLabel('import phantom icosahedron,\n half1 and half2')
        self.launchProtocol(prot5)
        self.volume = prot5.outputVolume
        self.volume.setOrigin(None)
        # The volume has no origin
        t = self.volume.getOrigin(force=True)
        x, y, z = t.getShifts()
        self.assertEqual(-90.0, x)
        self.assertEqual(-90.0, y)
        self.assertEqual(-90.0, z)
        # import half maps
        args = {'filesPath': volMapNamehalf1,
                'filesPattern': '',
                'samplingRate': _samplingRate
                }
        prot5_1 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot5_1.setObjLabel('import phantom\nicosahedron,\nhalf1')
        self.launchProtocol(prot5_1)
        self.half1 = prot5_1.outputVolume
        self.half1.setOrigin(None)
        # The volume has no origin
        t = self.volume.getOrigin(force=True)
        x, y, z = t.getShifts()
        self.assertEqual(-90.0, x)
        self.assertEqual(-90.0, y)
        self.assertEqual(-90.0, z)

        args = {'filesPath': volMapNamehalf2,
                'filesPattern': '',
                'samplingRate': _samplingRate
                }
        prot5_2 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot5_2.setObjLabel('import phantom\nicosahedron,\nhalf2')
        self.launchProtocol(prot5_2)
        self.half2 = prot5_2.outputVolume
        self.half2.setOrigin(None)
        # The volume has no origin
        t = self.volume.getOrigin(force=True)
        x, y, z = t.getShifts()
        self.assertEqual(-90.0, x)
        self.assertEqual(-90.0, y)
        self.assertEqual(-90.0, z)
        # import mask
        args = {'maskPath': volMaskName,
                'samplingRate': _samplingRate
                }
        prot6 = self.newProtocol(emprot.ProtImportMask, **args)
        prot6.setObjLabel('import mask')  # add origin to the mask
        self.launchProtocol(prot6)
        self.mask = prot6.outputMask
        self.mask.setOrigin(None)
        # The volume has no origin
        t = self.mask.getOrigin(force=True)
        x, y, z = t.getShifts()
        self.assertEqual(-90.0, x)
        self.assertEqual(-90.0, y)
        self.assertEqual(-90.0, z)

    def test_halfmaps_mask(self):
        """ If the input is a 3D map with half volumes associated
            Save them to the output directory
            Also create a mask and export it
        """
        # first create the 3 3dmaps and the mask
        self.import_volume_halfmaps()
        # run the export protocol
        from pwem import Domain
        XmippProtResolution3D = Domain.importFromPlugin('xmipp3.protocols',
                                                        'XmippProtResolution3D',
                                                        doRaise=True)

        prot = self.newProtocol(XmippProtResolution3D)
        prot.inputVolume.set(self.half1)
        prot.referenceVolume.set(self.half2)
        self.launchProtocol(prot)

        protExp = self.newProtocol(emprot.ProtExportDataBases)
        protExp.setObjLabel("export to DB\ndir3")
        protExp.exportVolume.set(self.volume)
        protExp.additionalVolumesToExport.set(True)
        protExp.exportAdditionalVolumes.set([self.protImportHalf1.outputVolume,
                                             self.protImportHalf2.outputVolume])
        protExp.exportFSC.set(prot.outputFSC)
        protExp.masksToExport.set(True)
        protExp.exportMasks.set([self.mask])
        protExp.exportAtomStruct.set(self._importAtomStructCIF())

        protExp.filesPath.set(os.getcwd() + "/dir3")
        self.launchProtocol(protExp)

        # Check the files were generated properly.
        # protExp._createFileNamesTemplates()
        dirName = protExp.filesPath.get()
        nameVolume = os.path.join(dirName, protExp.VOLUMENAME)
        nameFsc = os.path.join(dirName, "fsc_%02d.xml" % 1)
        nameAtomStruct = os.path.join(dirName, protExp.COORDINATEFILENAME)
        self.assertTrue(os.path.exists(nameVolume))
        self.assertTrue(os.path.exists(nameFsc))
        self.assertTrue(os.path.exists(nameAtomStruct))
        # assert results



# **************************************************************************
# *
# * Authors: Yunior C. Fonseca Reyna    (cfonseca@cnb.csic.es)
# *
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import os.path
import subprocess

from pwem.objects import EMSet
from pwem.viewers.viewers_data import RegistryViewerConfig
from pyworkflow import PYTHON
from pyworkflow.viewer import Viewer, View
from pwem.viewers.mdviewer.readers import SCIPION_PORT, SCIPION_OBJECT_ID


class MDView(View):

    def __init__(self, emSet: EMSet, protocol, port):
        self._emSet = emSet
        self.protocol = protocol
        self.port = port

    def show(self):
        env = os.environ
        env[SCIPION_PORT] = str(self.port)
        env[SCIPION_OBJECT_ID] = str(self._emSet.getObjId())
        visibleLabels = self.getVisibleLabels()

        subprocess.Popen(
            [PYTHON, "-m", "metadataviewer", "--extensionpath", os.path.join(os.path.dirname(__file__), "readers.py"),
            self._emSet.getFileName(), "--visiblelabels", visibleLabels])

    def getVisibleLabels(self):
        from pwem.viewers import VISIBLE
        config = RegistryViewerConfig.getConfig(type(self._emSet))
        if config is not None and VISIBLE in config:
            return config[VISIBLE]
        return ''


class MDViewer(Viewer):
    _name = 'Scipion metadata viewer'
    _targets = [EMSet]

    def _visualize(self, obj, **kwargs):
        return [MDView(obj, self.protocol, self._project.port)]







# **************************************************************************
# *
# * Authors:     R. Marabini (roberto@cnb.csic.es)
# *              Pablo Conesa (pconesa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" This module is to add protocols that need to be there for testing purposes
but nos available for regular users. This will hide them and prevent the their
 appearance in any GUI search or tree."""
import time

import pyworkflow.protocol as pwprot
import pyworkflow.protocol.params as params
from pyworkflow import VERSION_1_1


STRESS_NG = 'stress-ng'


class ProtTests(pwprot.Protocol):
    @classmethod
    def isDisabled(cls):
        """ Return True for all test protocols.
        Disabled protocols will not be offered in the available protocols."""
        return True


class ProtStress(ProtTests):
    """ stress  will  stress  test  a  computer system in various selectable
       ways. Several options require the program stress-ng.
    """
    _label = 'stress'
    _program = STRESS_NG
    _lastUpdateVersion = VERSION_1_1

    def __init__(self, **kwargs):
        ProtTests.__init__(self, **kwargs)
        self.xmippMic = {}

    # --------------------------- DEFINE param functions ----------------------

    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('noCpu', params.IntParam, default=0,
                      label="No. CPU stressors",
                      help="start N workers spinning on sqrt(rand())")
        form.addParam('noMem', params.IntParam, default=0,
                      label="No. memory stressors",
                      help="start N workers continuously calling mmap(2)/munmap"
                           "(2) and writing to the allocated memory. It will "
                           "use 100% per worker ")
        form.addParam('amountMem', params.IntParam, default=256,
                      label="Memory per stressors (Mb)",
                      help="allocate N bytes per vm worker. each stressors will"
                           " use 100% of a CPU")
        form.addParam('noIO', params.IntParam, default=0,
                      label="No IO stressors",
                      help="start N workers spinning on sync() (Disk io)")
        form.addParam('timeout', params.IntParam, default=60,
                      label="TimeOut (sec)",
                      help="timeout after N seconds. Total execution time is "
                           "timeout plus delay")
        form.addParam('delay', params.IntParam, default=0,
                      label="delay (sec)",
                      help="wait this seconds before stressing the system "
                           "seconds")
        form.addParam('extraParams', params.StringParam,
                      label='Additional parameters', default=" ",
                      help='Additional parameters for stress-ng '
                           '(http://kernel.ubuntu.com/~cking/stress-ng/)')

    # ---------------------- INSERT steps functions ----------------------------

    def _insertAllSteps(self):
        if self.delay != 0:
            self._insertFunctionStep('delayStep')
        self._insertFunctionStep('stressStep')

    # ------------------------ STEPS functions ---------------------------------
    def delayStep(self):
        time.sleep(self.delay)

    def stressStep(self):
        if self.noCpu == 0 and self.noIO == 0 and self.noMem == 0:
            time.sleep(self.timeout)
        else:
            args = " "
            if self.noCpu != 0:
                args += " --cpu %d" % self.noCpu
            if self.noIO != 0:
                args += " --io %d" % self.noIO
            if self.noMem != 0:
                args += " --mmap %d" % self.noMem
            if self.amountMem != 0:
                args += " --mmap-bytes %dM" % self.amountMem
            args += " --timeout %ds" % self.timeout
            args += " --metrics-brief"
            if self.extraParams is not None:
                args += " %s " % self.extraParams
            self.runJob(self._program, args)

    # ------------------------ INFO functions ----------------------------------
    def _validate(self):
        message = []
        from shutil import which
        if self.noCpu == 0 and self.noIO == 0 and self.noMem == 0:
            pass
        else:
            if which(self._program) is None:
                message = ["Cannot find executable %s" % self._program]
        return message

    def _summary(self):
        message = "%d CPU, %d Mem, %d IO stressor" % (self.cpu, self.mem,
                                                      self.io)
        return [message]

    def _methods(self):
        return []


# class ProtOutputTest(ProtTests):
#     """ Protocol to test scalar output and input linking"""
#     _label = 'test output'
#
#     def __init__(self, **args):
#         Protocol.__init__(self, **args)
#         self.name = params.String(args.get('name', None))
#
#     def _defineParams(self, form):
#
#         section = form.addSection("Input")
#         section.addParam('iBoxSize', params.IntParam, allowsPointers=True,
#                          default=10,
#                          label='Input box size as Integer',
#                          validators=[params.Positive])
#
#         section.addParam('nullableInteger', params.IntParam, allowsPointers=True,
#                          label='Nullable Integer', allowsNull=True)
#
#     def _createOutputStep(self):
#         # New Output would be an Integer
#         boxSize = Integer(10)
#
#         if self.iBoxSize.hasValue():
#             boxSize.set(2*int(self.iBoxSize.get()))
#
#         self._defineOutputs(oBoxSize=boxSize)
#
#     def _insertAllSteps(self):
#         self._insertFunctionStep('_createOutputStep')
#
#
# class ProtMultiPointerTest(ProtTests):
#     """ Class to test how multipointer params are exported to json"""
#     def _defineParams(self, form):
#
#         # This should cover Multipointer params that points to attributes...
#         # therefore extended attribute of pointers should be used
#         form.addParam('mpToAttr', params.MultiPointerParam,
#                       label="Multipointer to attribute",
#                       pointerClass='String',
#                       help="Should point to String inside another protocol")
#
#         # This should cover Multipointer params that points to protocols...
#         # therefore extended attribute of pointers should NOT be used
#         form.addParam('mpToProts', params.MultiPointerParam,
#                       label="Multipointer to sets",
#                       pointerClass='Protocol',
#                       help="Should point to another protocol")




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import join, basename, exists
from collections import OrderedDict

from pyworkflow.utils import findRootFrom

import pwem.objects as emobj
import pwem.constants as emcts


class ScipionImport:
    """ Import 
    """
    def __init__(self, protocol, sqliteFile):
        self.protocol = protocol
        self._sqliteFile = sqliteFile
        self.copyOrLink = protocol.getCopyOrLink()
    
    def importMicrographs(self):
        """ Import a SetOfMicrographs from a given sqlite file. """
        inputSet = emobj.SetOfMicrographs(filename=self._sqliteFile)
        self._findImagesPath(inputSet)
        
        micSet = self.protocol._createSetOfMicrographs()
        micSet.setObjComment('Micrographs imported from sqlite file:\n%s'
                             % self._sqliteFile)

        # Update both samplingRate and acquisition with parameters
        # selected in the protocol form.
        # Acquisition should be updated before, just to ensure that
        # scannedPixedSize will be computed properly when calling
        # setSamplingRate
        self.protocol.fillAcquisition(micSet.getAcquisition())
        self.protocol.setSamplingRate(micSet)
        micSet.setIsPhaseFlipped(self.protocol.haveDataBeenPhaseFlipped.get())
        # Read the micrographs from the 'self._sqliteFile' metadata
        # but fixing the filenames with new ones (linked or copy to extraDir)
        micSet.copyItems(inputSet, updateItemCallback=self._updateParticle)
        self.protocol._defineOutputs(outputMicrographs=micSet)

    def importCoordinates(self, inputMics):
        """ Import a SetOfCoordinates from a given sqlite file. """
        inputSet = emobj.SetOfCoordinates(filename=self._sqliteFile)
        # self._findImagesPath(inputSet)
        
        coorSet = self.protocol._createSetOfCoordinates(inputMics)
        coorSet.copyInfo(inputSet)
        coorSet.setObjComment('Coordinates imported from sqlite file:\n%s'
                             % self._sqliteFile)

        # Update both samplingRate and acquisition with parameters
        # selected in the protocol form.
        # Acquisition should be updated before, just to ensure that
        # scannedPixedSize will be computed properly when calling
        # setSamplingRate
        coorSet.setBoxSize(self.protocol.boxSize.get())
        # Read the micrographs from the 'self._sqliteFile' metadata
        # but fixing the filenames with new ones (linked or copy to extraDir)
        coorSet.copyItems(inputSet)
        self.protocol._defineOutputs(outputCoordinates=coorSet)
        self.protocol._defineSourceRelation(inputMics, coorSet)

    def importParticles(self):
        """ Import particles from a metadata 'images.xmd' """
        inputSet = emobj.SetOfParticles(filename=self._sqliteFile)
        inputSet.loadProperty('_alignment', emcts.ALIGN_NONE)
        inputSet.loadProperty('_hasCtf', False)
        self._findImagesPath(inputSet)

        partSet = self.protocol._createSetOfParticles()
        partSet.copyInfo(inputSet)
        
        partSet.setObjComment('Particles imported from Scipion sqlite file:\n%s'
                              % self._sqliteFile)
        partSet.copyItems(inputSet, updateItemCallback=self._updateParticle)
        
        # Update both samplingRate and acquisition with parameters
        # selected in the protocol form
        self.protocol.setSamplingRate(partSet)
        self.protocol.fillAcquisition(partSet.getAcquisition())
        partSet.setIsPhaseFlipped(self.protocol.haveDataBeenPhaseFlipped.get())
        self.protocol._defineOutputs(outputParticles=partSet)
        
    def _findImagesPath(self, inputSet):
        """ Find the relative path from which the images exists
        revative to the sqlite location.
        """
        # Store which images stack have been linked/copied and the new path
        self._imgDict = {}
        img = inputSet.getFirstItem()
        self._imgPath = findRootFrom(self._sqliteFile, img.getFileName())
        
        if self._imgPath is None:
            self.protocol.warning("Binary data was not found from sqlite: %s"
                                  % self._sqliteFile)
    
    def validate(self):
        """ Try to find errors on import. """
        errors = []
        return errors
        
    def validateMicrographs(self):
        return self.validate()
    
    def validateParticles(self):
        return self.validate()
        
    def _updateParticle(self, particle, partRow):
        if self._imgPath:
            # Create a link or copy files to extraPath
            # and update the Row properly
            fn = particle.getFileName()
            imgBase = basename(fn)
            imgDst = self.protocol._getExtraPath(imgBase)
            if not exists(imgDst):
                self.copyOrLink(join(self._imgPath, fn), imgDst)
            particle.setFileName(imgDst)

    def loadAcquisitionInfo(self):
        """ Return a dictionary with acquisition values and 
        the sampling rate information.
        In the case of Scipion, these values will be read from the
        'Properties' table of the particles.sqlite file.
        """
        acq = OrderedDict()
        inputSet = emobj.SetOfParticles(filename=self._sqliteFile)

        def _get(key):
            return inputSet.getProperty(key)
        acq['samplingRate'] = _get('_samplingRate')
        acq['voltage'] = _get('_acquisition._voltage')
        acq['amplitudeContrast'] = _get('_acquisition._amplitudeContrast')
        acq['sphericalAberration'] = float(_get('_acquisition._sphericalAberration'))

        return acq

    def getMicCTF(self, mic):
        """ Retrieve the CTF associated to this micrograph. """
        # The first time this function is called, the set of micrographs
        # will be loaded from the given sqlite file
        if not hasattr(self, 'ctfSet'):
            self.ctfSet = emobj.SetOfCTF(filename=self._sqliteFile)

        return self.ctfSet[mic.getObjId()]






#!/usr/bin/env python

import sys
from pwem.viewers import Chimera


def main():
    fileName = sys.argv[1]

    # Clean filename in case it comes annotated like 1@xxxmrc:mrcs
    fileName = fileName.split(":")[0]

    # if contains @
    if "@" in fileName:
        fileName = fileName.split("@")[1]

    Chimera().runProgram(args=fileName)


if __name__ == '__main__':
    main()



# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This modules contains file handlers to be registered in the scipion file browser
"""
import os

import pwem
import pyworkflow.utils as pwutils
from pwem.convert import Ccp4Header
from pyworkflow import gui
from pyworkflow.gui.browser import FileHandler, isStandardImage

from pwem import emlib

import logging
logger = logging.getLogger(__name__)


class ImageFileHandler(FileHandler):
    _image = emlib.Image()
    _index = ''

    def _getImageString(self, filename):
        if isStandardImage(filename):
            return "Image file."
        x, y, z, n = emlib.getImageSize(filename)
        objType = 'Image'
        objFileType = "link" if os.path.islink(filename) else "file"
        dimMsg = "*%(objType)s %(objFileType)s*\n  dimensions: %(x)d x %(y)d"
        expMsg = "Columns x Rows "
        if z > 1:
            dimMsg += " x %(z)d"
            expMsg += " x Slices"
            objType = 'Volume'
        if n > 1:
            dimMsg += " x %(n)d"
            expMsg += " x Objects"
            objType = 'Stack'

        if Ccp4Header.isCompatible(filename):
            ccp4Reader= Ccp4Header(filename, readHeader=True)
            expMsg += "\nHeader info: \nSampling rate: %1.2f, %1.2f, %1.2f" % ccp4Reader.getSampling()
            expMsg += "\nOrigin: %1.2f, %1.2f, %1.2f" % ccp4Reader.getOrigin()


        return (dimMsg + "\n" + expMsg) % locals()

    def _getImagePreview(self, filename):

        # If file size if big
        if pwutils.getFileSize(filename) > (pwem.Config.MAX_PREVIEW_FILE_SIZE * 1024*1024):
            return

        dim = 128

        if isStandardImage(filename):
            self.tkImg = gui.getImage(os.path.abspath(filename),
                                      tkImage=True, maxheight=dim)
        else:
            fn = self._index + filename
            self.tkImg = getTkImage(self._image, fn, dim)

        return self.tkImg

    def getFilePreview(self, objFile):
        fn = objFile.getPath()
        return self._getImagePreview(fn), self._getImageString(fn)

    def getFileActions(self, objFile):
        from .views import DataView
        fn = objFile.getPath()
        return [('Open with Xmipp viewer', lambda: DataView(fn).show(),
                 pwutils.Icon.ACTION_VISUALIZE)]


class ParticleFileHandler(ImageFileHandler):
    def getFileIcon(self, objFile):
        return 'file_image.gif' if not objFile.isLink() else 'file_image_link.gif'


class VolFileHandler(ImageFileHandler):
    def getFileIcon(self, objFile):
        return 'file_vol.gif'

class StackHandler(ImageFileHandler):
    _index = '1@'

    def getFileIcon(self, objFile):
        return 'file_stack.gif' if not objFile.isLink() else 'file_stack_link.gif'

class ImajeJFileHandler(FileHandler):

    def getFileActions(self, objFile):
        from pyworkflow.viewer import CommandView

        fn = objFile.getPath()
        fijiPath = pwem.Config.IMAGEJ_BINARY_PATH

        if fijiPath:
            cmd = '%s "%s"' % (fijiPath, fn)
            return [('Open with ImageJ/Fiji', lambda: CommandView(cmd).show(),
                 pwutils.Icon.ACTION_VISUALIZE)]

    def getFileIcon(self, objFile):
        return 'file_image.gif' if not objFile.isLink() else 'file_image_link.gif'


class ChimeraHandler(FileHandler):

    def getFileActions(self, objFile):
        from .viewer_chimera import ChimeraView
        fn = objFile.getPath()
        return [('Open with Chimera', lambda: ChimeraView(fn).show(),
                 pwutils.Icon.ACTION_VISUALIZE)]

    def getFileIcon(self, objFile):
        return 'file_text.gif'


class MdFileHandler(ImageFileHandler):
    def getFileIcon(self, objFile):
        if objFile.isLink():
            return 'file_md_link.gif'
        else:
            return 'file_md.gif'

    def _getImgPath(self, mdFn, imgFn):
        """ Get ups and ups until finding the relative location to images. """
        path = os.path.dirname(mdFn)
        index, fn = emlib.FileName(imgFn).decompose()

        while path and path != '/':
            newFn = os.path.join(path, fn)
            if os.path.exists(newFn):
                if index:
                    newFn = '%d@%s' % (index, newFn)
                return newFn
            path = os.path.dirname(path)

        return None

    def _getMdString(self, filename, block=None):
        md = emlib.MetaData()
        if block:
            md.read(block + '@' + filename)
        else:
            md.read(filename, 1)
        labels = md.getActiveLabels()
        msg = "Metadata items: *%d*\n" % md.getParsedLines()
        msg += "Metadata labels: " + ''.join(
            ["\n   - %s" % emlib.label2Str(l)
             for l in labels])

        imgPath = None
        for label in labels:
            if emlib.labelIsImage(label):
                imgPath = self._getImgPath(filename,
                                           md.getValue(label, md.firstObject()))
                if imgPath is None or not os.path.exists(imgPath):
                    imgPath = None
                break

        # If there is an image and is not too big
        if imgPath and pwutils.getFileSize(imgPath) < (pwem.Config.MAX_PREVIEW_FILE_SIZE * 1024*1024):

            self._imgPreview = self._getImagePreview(imgPath)
            self._imgInfo = self._getImageString(imgPath)
        return msg

    def getFilePreview(self, objFile):
        self._imgPreview = None
        self._imgInfo = None
        filename = objFile.getPath()
        ext = pwutils.getExt(filename)

        if ext == '.xmd' or ext == '.ctfparam' or ext == '.pos' or ext == '.doc':
            msg = "*Metadata File* "
            blocks = emlib.getBlocksInMetaDataFile(filename)
            nblocks = len(blocks)
            if nblocks <= 1:
                mdStr = self._getMdString(filename)
                msg += "  (single block)\n"
                if self._imgInfo:
                    msg += "\nFirst item: \n" + self._imgInfo
                msg += '\n' + mdStr
            else:
                mdStr = self._getMdString(filename, blocks[0])
                msg += "  (%d blocks) " % nblocks
                if self._imgInfo:
                    msg += "\nFirst item: \n" + self._imgInfo
                msg += "\nFirst block: \n" + mdStr
                msg += "\nAll blocks:" + ''.join(
                    ["\n  - %s" % b for b in blocks])
        elif ext == '.star':
            msg = "*Relion STAR file* \n"
            msg += self._getMdString(filename)

        return self._imgPreview, msg


# Image methods for filehandlers
def getPILImage(imageXmipp, dim=None, normalize=True):
    """ Given an image read by Xmipp, convert it to PIL. """
    from PIL import Image

    if normalize:
        imageXmipp.convert2DataType(emlib.DT_UCHAR, emlib.CW_ADJUST)

    imageData = imageXmipp.getData()
    image = Image.fromarray(imageData)
    if dim:
        size = int(dim), int(dim)
        image.thumbnail(size, Image.ANTIALIAS)
    return image


def getTkImage(imageXmipp, filename, dim):
    from PIL import ImageTk
    imageXmipp.readPreview(filename, dim)
    return ImageTk.PhotoImage(getPILImage(imageXmipp))


def getImageFromPath(imagePath):
    """ Read an image using Xmipp, convert to PIL
    and then return as expected by Tk.
    """
    img = emlib.Image(imagePath)
    imgPIL = getPILImage(img)
    from PIL import ImageTk
    imgTk = ImageTk.PhotoImage(imgPIL)

    return imgTk



import os
import time

from pwem.objects import SetOfMovies
from pwem.protocols import EMProtocol
from pyworkflow import BETA
from pyworkflow.protocol import params
import pyworkflow.utils as pwutils

class ProtMovieEraser(EMProtocol):
    """ It will REMOVE movies that already have been aligned into micrographs.
    WARNING: There is no way back. Be sure you understand the consequences."""
    _label = "movie eraser"
    _devStatus = BETA
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.deletedMovies = []

    def _defineParams(self, form):

        form.addSection(label='Input')
        form.addParam('inputMicrographs', params.PointerParam,
                      pointerClass='SetOfMicrographs',
                      label=pwutils.Message.LABEL_INPUT_MIC, important=True,
                      help='Select the SetOfMicrographs to drive the movie DELETION process.')

        form.addParam('movieSet', params.PointerParam,
                      pointerClass='SetOfMovies',
                      label="Movies source", important=True,
                      help='Select the SetOfMovies to locate the movie files. Match will be done by ids!.')

        form.addParam("dryMode", params.BooleanParam, label="Dry mode", default=True, expertLevel=params.LEVEL_ADVANCED)

    # -------------------------- STEP methods ---------------------------------
    def _insertAllSteps(self):

        self._insertFunctionStep(self.deleteMoviesStep)

    def deleteMoviesStep(self):
        """ Delete all movies origin of the micset until set is closed"""
        lastMicId = 0
        while not self._isAllDone():
            # For each mic not done yet
            for mic in self._getMicSet().iterItems(where="id > %s" % lastMicId):
                self._deleteMovieFromMic(mic)
                lastMicId = mic.getObjId()

            # wait some time
            time.sleep(30)


    def _deleteMovieFromMic(self, mic):
        """ Deletes the movie source of the mic object"""
        movieFile = self._getMovieFileFromMic(mic)

        # Check existence
        if not os.path.exists(movieFile):
            print("%s is gone already!." % movieFile)

        # if dry mode
        elif self.dryMode.get():
            print("%s would have been deleted" % movieFile, flush=True)

        # Delete!!!
        else:
            # Delete the move
            os.remove(movieFile)
            print("%s deleted." % movieFile, flush=True)

        # Annotate it
        self.deletedMovies.append(movieFile)

    def _getMovieFileFromMic(self, mic):
        """ Returns the movie path that was used in the micrograph movie alignment"""
        movieSet = self._getMovieSet()

        # To match the movie use the ID:
        # movie alignment protocols are copying the movie id to the mic --> mic.copyObjId(movie)
        movie = movieSet[mic.getObjId()]

        movieFile = movie.getFileName()

        if os.path.islink(movieFile):
            movieFile = os.path.realpath(movieFile)

        return movieFile

    def _getMovieSet(self):
        """ Returns the SetOFMovies related to the input set of micrographs"""
        # This is not possible since the relations are stored in the run.db of the input protocol
        # and there is no easy way to get them from this run
        # parents= self.getProject().getSourceParents(self._getMicSet())
        #
        # # Get the parent protocols
        # for parent in parents:
        #     # We need to update them taking the most recent data form its own run.db
        #     self.getProject()._updateProtocol(parent)
        #     for attr, output in parent.iterOutputAttributes():
        #         if isinstance(output, SetOfMovies):
        #             return output
        #
        # return None
        return self.movieSet.get()

    def _getMicSet(self):
        """ Returns te input set of micrographs"""
        mics = self.inputMicrographs.get()
        mics.loadAllProperties()
        return mics

    def _isAllDone(self):
        """ Returns True if all work is done. All movies deleted """
        mics = self._getMicSet()
        return mics.isStreamClosed() and mics.getSize() == len(self.deletedMovies)

    @classmethod
    def worksInStreaming(cls):
        return True
    # -------------------------- INFO functions -------------------------------
    def _warnings(self):
        if not self.dryMode.get():
            return ["If you continue, all movies related to the set of micrographs will be DELETED!. Try to run it in \"dry mode\" to double check first."]

    def _summary(self):
        summary = []
        summary.append("Dry mode is active. Harmless" if self.dryMode.get() else "Dry mode deactivated, harm can be done!")
        return summary




# ***************************************************************************
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/

import pyworkflow.tests as pwtests

import pwem.protocols as emprot
from pyworkflow.object import Pointer


class TestImportBase(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dsXmipp = pwtests.DataSet.getDataSet('xmipp_tutorial')
        cls.dsEmx = pwtests.DataSet.getDataSet('emx')
        cls.dsMda = pwtests.DataSet.getDataSet('mda')
        cls.dsRelion = pwtests.DataSet.getDataSet('relion_tutorial')
        
    def checkOutput(self, prot, outputName, conditions=[]):
        """ Check that an output was generated and
        the condition is valid. 
        """
        o = getattr(prot, outputName, None)
        locals()[outputName] = o 
        self.assertIsNotNone(o, "Output: %s is None" % outputName)
        for cond in conditions:
            self.assertTrue(eval(cond), 'Condition failed: ' + cond)
        
    
class TestImportParticles(TestImportBase):

    def test_pattern(self):
        """ Import several Particles from a given pattern.
        """
        args = {'importFrom': emprot.ProtImportParticles.IMPORT_FROM_FILES,
                'filesPath': self.dsXmipp.getFile('particles/'),
                'filesPattern': 'BPV_????_ptcls.hdf',
                'amplitudeConstrast': 0.1,
                'sphericalAberration': 2.,
                'voltage': 100,
                'samplingRate': 2.1,
                'haveDataBeenPhaseFlipped': True
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        protMicImport = self.newProtocol(emprot.ProtImportParticles, **args)
        protMicImport.setObjLabel('from files')
        self.launchProtocol(protMicImport)

        # Id's should be taken from filename    
        args['filesPattern'] = 'BPV_####_ptcls.hdf'
        protMicImport = self.newProtocol(emprot.ProtImportParticles, **args)
        protMicImport.setObjLabel('from files (with mic id)')
        self.launchProtocol(protMicImport)

    def test_fromEmx(self):
        """ Import an EMX file with Particles and defocus
        """
        args = {'importFrom': emprot.ProtImportParticles.IMPORT_FROM_EMX,
                'amplitudeConstrast': 0.1,
                'sphericalAberration': 2.,
                'voltage': 100,
                'alignType': 3,
                'samplingRate': 2.46,
                'magnification': 10000,
                'haveDataBeenPhaseFlipped': True
                }
        args['emxFile'] = self.dsEmx.getFile('particles/particles.emx')
        protEmxImport = self.newProtocol(emprot.ProtImportParticles, **args)
        protEmxImport.setObjLabel('from emx (particles)')
        self.launchProtocol(protEmxImport)

        # Import some Particles from EMX
        args['emxFile'] = self.dsEmx.getFile('coordinatesT1')
        protEmxImport = self.newProtocol(emprot.ProtImportParticles, **args)
        protEmxImport.setObjLabel('from emx (with coords)')
        self.launchProtocol(protEmxImport)

    def test_fromXmipp(self):
        """ Import an Xmipp file with Particles and defocus
        """
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_XMIPP3,
                                 mdFile=self.dsXmipp.getFile('gold/xmipp_ml2d_images.xmd'),
                                 magnification=10000,
                                 samplingRate=1,
                                 haveDataBeenPhaseFlipped=True
                                 )
        prot1.setObjLabel('from xmipp (ml2d)')
        self.launchProtocol(prot1)

    def test_fromXmippWithMic(self):
        """ Import an EMX file with Particles and defocus
        """
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_XMIPP3,
                                 mdFile=self.dsXmipp.getFile('gold/images10.xmd'),
                                 magnification=10000,
                                 samplingRate=1,
                                 haveDataBeenPhaseFlipped=True
                                 )
        prot1.setObjLabel('from xmipp (with mic id)')
        self.launchProtocol(prot1)

    def test_fromRelionRefine3D(self):
        """ Import an EMX file with Particles and defocus
        """
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 objLabel='from relion (auto-refine 3d)',
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_RELION,
                                 starFile=self.dsRelion.getFile('import/refine3d/extra/relion_it025_data.star'),
                                 magnification=10000,
                                 samplingRate=7.08,
                                 haveDataBeenPhaseFlipped=True
                                 )
        self.launchProtocol(prot1)
        self.checkOutput(prot1, 'outputParticles', ['outputParticles.hasAlignmentProj()',
                                                    'outputParticles.isPhaseFlipped()'])
        
    def test_fromRelionClassify2D(self):
        """ Import an EMX file with Particles and defocus
        """
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 objLabel='from relion (classify 2d)',
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_RELION,
                                 starFile=self.dsRelion.getFile('import/classify2d/extra/relion_it015_data.star'),
                                 magnification=10000,
                                 samplingRate=7.08,
                                 haveDataBeenPhaseFlipped=True
                                 )
        self.launchProtocol(prot1)
        self.checkOutput(prot1, 'outputParticles', ['outputParticles.hasAlignment2D()',
                                                    'outputParticles.isPhaseFlipped()'])
        self.checkOutput(prot1, 'outputClasses')


        # Add tests for classes selector. representative
        classSelector = self.newProtocol(emprot.ProtClassesSelector,
                                         objLabel='representatives from 2 mayor classes',
                                         firstNElements=2,
                                         extractRepresentative=True
                                         )
        classSelector.inputClasses = Pointer(prot1, extended='outputClasses')
        self.launchProtocol(classSelector)

        self.assertSetSize(classSelector.output, size=2)

        # Add tests for classes selector: items
        classSelector2 = self.newProtocol(emprot.ProtClassesSelector,
                                         objLabel='items from 3 mayor classes',
                                         firstNElements=3,
                                         extractRepresentative=False
                                         )
        classSelector2.inputClasses = Pointer(prot1, extended='outputClasses')
        self.launchProtocol(classSelector2)

        self.assertSetSize(classSelector2.output, size=1739)


    def test_fromRelionClassify3D(self):
        """ Import an EMX file with Particles and defocus
        """
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 objLabel='from relion (classify 3d)',
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_RELION,
                                 starFile=self.dsRelion.getFile('import/classify3d/extra/relion_it015_data.star'),
                                 magnification=10000,
                                 samplingRate=7.08,
                                 haveDataBeenPhaseFlipped=True
                                 )
        self.launchProtocol(prot1)         
        
    def test_fromScipionReconstruct(self):
        """ Import the particles with 3D projection directions for reconstruct
        a volume. Actually this test use a similar .sqlite file than
        the output of
        test_fromRelionRefine3D
        """
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 objLabel='from scipion (to-reconstruct)',
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_SCIPION,
                                 sqliteFile=self.dsRelion.getFile('import/case2/particles.sqlite'),
                                 magnification=10000,
                                 samplingRate=7.08,
                                 haveDataBeenPhaseFlipped=True
                                 )
        self.launchProtocol(prot1)
        
        self.checkOutput(prot1, 'outputParticles', ['outputParticles.hasCTF()', 
                                                    'outputParticles.hasAlignmentProj()',
                                                    'outputParticles.isPhaseFlipped()'])
        
    def test_fromImagic(self):
        """ Import particles from imagic stack.
        """
        args = {'importFrom': emprot.ProtImportParticles.IMPORT_FROM_FILES,
                'filesPath': self.dsMda.getFile('particles/'),
                'filesPattern': 'particles.hed',
                'amplitudeConstrast': 0.1,
                'sphericalAberration': 2.,
                'voltage': 100,
                'samplingRate': 2.1,
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        protImport = self.newProtocol(emprot.ProtImportParticles, **args)
        protImport.setObjLabel('from imagic')
        self.launchProtocol(protImport)
        
        self.checkOutput(protImport, 'outputParticles', [])



from unittest.mock import MagicMock

from pwem.objects import CTFModel, Micrograph, Particle, Coordinate


def getMicList(start=1, end=3):
    """ Mocks a mic dict"""

    newMicDict = {}

    for index in range(start, end + 1):
        mic = Micrograph()
        mic.setMicName("mic%s" % index)
        newMicDict[index] = mic

    return newMicDict


def getSoMMock(start=1, end=3, name=None):
    """ Mocks a set of micrographs"""
    som = MagicMock(name=name)
    micList = getMicList(start, end)
    som.__iter__.return_value = micList.values()
    som.iterItems.return_value = list(micList.values())
    som.items.side_effect = micList.items
    som.__getitem__.side_effect = lambda key: micList.__getitem__(key)
    return som


def getCTFList(start=1, end=3):
    """ Mocks a CTF dict"""
    newCtfDict = {}

    for index in range(start, end + 1):
        ctf = CTFModel()
        mic = Micrograph()
        mic.setMicName("mic%s" % index)
        ctf.setMicrograph(mic)
        newCtfDict[mic.getMicName()] = ctf

    return newCtfDict


def getSoCTFsMock(start=1, end=3):
    """ Mocks a set of CTFs"""
    som = MagicMock()
    som.__iter__.return_value = getCTFList(start, end).values()
    return som


def getParticleList(start=1, end=3):
    """ Mocks a Particle dict"""
    newPartDict = {}

    for index in range(start, end + 1):
        part = Particle()
        coord = Coordinate()
        coord.setMicId(index)
        coord.setMicName(getMicNameFromId(index))
        part.setCoordinate(coord)
        newPartDict[index] = part

    return newPartDict


def getMicNameFromId(id):
    return "mic%s" % id


def getSoPartMock(start=1, end=3, name=None):
    """ Mocks a set of Particles"""
    som = MagicMock(name=name)
    som.__iter__.return_value = getParticleList(start, end).values()
    return som


def get2DCoordList(start=1, end=3):
    """ Mocks a 2d coordinates dict"""

    newCoordDict = {}

    for index in range(start, end + 1):
        coord = Coordinate()
        newCoordDict[index] = coord

    return newCoordDict


def getSo2DCMock(start=1, end=3):
    """ Mocks a set of 2D coordinates"""
    som = MagicMock()
    som.__iter__.return_value = get2DCoordList(start, end).values()
    return som



# **************************************************************************
# *
# * Authors:     Josue Gomez Blanco (josue.gomez-blanco@mcgill.ca)
# *              Roberto Marabini (roberto@cnb.csic.es)
# *              Airen Zaldivar Peraza (azaldivar@cnb.csic.es)
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk)
# *
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import enum
from os.path import exists, getmtime
from datetime import datetime
from collections import OrderedDict

import pyworkflow.object as pwobj
import pyworkflow.protocol.constants as pwcts
import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils

import pwem.objects as emobj

from .protocol import EMProtocol


class ProtMicrographs(EMProtocol):
    pass


class ProtCTFMicOutputs(enum.Enum):
    outputCTF = emobj.SetOfCTF


class ProtCTFMicrographs(ProtMicrographs):
    """ Base class for all protocols that estimates the CTF"""

    _possibleOutputs = ProtCTFMicOutputs

    def __init__(self, **kwargs):
        EMProtocol.__init__(self, **kwargs)
        self.stepsExecutionMode = pwcts.STEPS_PARALLEL
        self.isFirstTime = pwobj.Boolean(False)

    # -------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label=pwutils.Message.LABEL_CTF_ESTI)
        form.addParam('recalculate', params.BooleanParam, default=False,
                      condition='recalculate',
                      label="Do recalculate ctf?")

        form.addParam('continueRun', params.PointerParam, allowsNull=True,
                      condition='recalculate', label="Input previous run",
                      pointerClass=self.getClassName())
        form.addHidden('sqliteFile', params.FileParam, condition='recalculate',
                       allowsNull=True)

        form.addParam('inputMicrographs', params.PointerParam, important=True,
                      condition='not recalculate',
                      label=pwutils.Message.LABEL_INPUT_MIC,
                      pointerClass='SetOfMicrographs')

        form.addParam('AutoDownsampling', params.BooleanParam, default=False,
                      label='Automatic Downsampling Factor',
                      help='Recommended value to downsample')

        form.addParam('ctfDownFactor', params.FloatParam, default=1.,
                      label='Manual CTF Downsampling factor',
                      condition='not AutoDownsampling',  # 'not recalculate',
                      help='Set to 1 for no downsampling. Non-integer downsample '
                           'factors are possible. This downsampling is only used '
                           'for estimating the CTF and it does not affect any '
                           'further calculation. Ideally the estimation of the '
                           'CTF is optimal when the Thon rings are not too '
                           'concentrated at the origin (too small to be seen) '
                           'and not occupying the whole power spectrum (since '
                           'this downsampling might entail aliasing).')

        self._defineProcessParams(form)

        line = form.addLine('Resolution', condition='not recalculate',
                            help='Give a value in digital frequency '
                                 '(i.e. between 0.0 and 0.5). These cut-offs '
                                 'prevent the typical peak at the center of the'
                                 ' PSD and high-resolution terms where only '
                                 'noise exists, to interfere with CTF '
                                 'estimation. The default lowest value is 0.05 '
                                 'but for micrographs with a very fine sampling '
                                 'this may be lowered towards 0. The default '
                                 'highest value is 0.35, but it should be '
                                 'increased for micrographs with signals '
                                 'extending beyond this value. However, if '
                                 'your micrographs extend further than 0.35, '
                                 'you should consider sampling them at a finer '
                                 'rate.')
        line.addParam('lowRes', params.FloatParam, default=0.05, label='Lowest')
        line.addParam('highRes', params.FloatParam, default=0.35, label='Highest')
        line = form.addLine('Defocus search range (microns)',
                            condition='not recalculate',
                            expertLevel=pwcts.LEVEL_ADVANCED,
                            help='Select _minimum_ and _maximum_ values for '
                                 'defocus search range (in microns). Underfocus'
                                 ' is represented by a positive number.')
        line.addParam('minDefocus', params.FloatParam, default=0.25,
                      label='Min')
        line.addParam('maxDefocus', params.FloatParam, default=4.,
                      label='Max')

        form.addParam('windowSize', params.IntParam, default=512,
                      expertLevel=pwcts.LEVEL_ADVANCED,
                      label='Window size', condition='not recalculate',
                      help='The PSD is estimated from small patches of this '
                           'size. Bigger patches allow identifying more '
                           'details. However, since there are fewer windows, '
                           'estimations are noisier.')

        form.addParallelSection(threads=2, mpi=1)

    def _defineProcessParams(self, form):
        """ This method should be implemented by subclasses
        to add other parameter relatives to the specific operation."""
        pass

    # -------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        """ Insert the steps to perform CTF estimation, or re-estimation,
        on a set of micrographs.
        """
        self._defineCtfParamsDict()
        self.micDict = OrderedDict()

        if not self.recalculate:
            self.initialIds = self._insertInitialSteps()
            micDict, self.streamClosed = self._loadInputList()
            ctfIds = self._insertNewMicsSteps(micDict.values())
            self._insertFinalSteps(ctfIds)
            # For the streaming mode, the steps function have a 'wait' flag
            # that can be turned on/off. For example, here we insert the
            # createOutputStep but it wait=True, which means that can not be
            # executed until it is set to False
            # (when the input micrographs stream is closed)
            waitCondition = self._getFirstJoinStepName() == 'createOutputStep'
        else:
            if self.isFirstTime:
                # Insert previous estimation or re-estimation an so on...
                self._insertPreviousSteps()
                self.isFirstTime.set(False)
            ctfIds = self._insertRecalculateSteps()
            # For now the streaming is not allowed for recalculate CTF
            waitCondition = False

        self._insertFunctionStep('createOutputStep', prerequisites=ctfIds,
                                 wait=waitCondition)

    def _insertInitialSteps(self):
        """ Override this function to insert some steps before the
        estimate ctfs steps.
        Should return a list of ids of the initial steps. """

        pwutils.makePath(self._getExtraPath('DONE'))
        return []

    def _insertNewMicsSteps(self, inputMics):
        """ Insert steps to process new mics (from streaming)
        Params:
            inputMics: input mics set to be check
        """
        return self._insertNewMics(inputMics,
                                   lambda mic: mic.getMicName(),
                                   self._insertCtfStep,
                                   self._insertCtfListStep,
                                   *self._getCtfArgs())

    def _insertRecalculateSteps(self):
        recalDeps = []
        # For each psd insert the steps to process it
        self.recalculateSet = emobj.SetOfCTF(filename=self.sqliteFile.get(),
                                             objDoStore=False)
        inputMics = self.getInputMicrographs()
        for ctf in self.recalculateSet:
            line = ctf.getObjComment()
            if ctf.isEnabled() and line:
                # CTF Re-estimation
                # Make estimation steps independent between them
                objId = ctf.getObjId()
                stepId = self._insertFunctionStep('reEstimateCtfStep',
                                                  objId, prerequisites=[])
                recalDeps.append(stepId)
                self.micDict[objId] = inputMics[objId].clone()
        return recalDeps

    def _insertFinalSteps(self, deps):
        """ This should be implemented in subclasses"""
        return deps

    def _getFirstJoinStepName(self):
        # This function will be used for streaming, to check which is
        # the first function that need to wait for all micrographs
        # to have completed, this can be overwritten in subclasses
        # (e.g., in Xmipp 'sortPSDStep')
        return 'createOutputStep'

    def _getFirstJoinStep(self):
        for s in self._steps:
            if s.funcName == self._getFirstJoinStepName():
                return s
        return None

    # -------------------------- STEPS functions ------------------------------
    def _insertCtfStep(self, mic, prerequisites, *args):
        """ Basic method to insert an estimation step for a given micrograph. """
        micStepId = self._insertFunctionStep('estimateCtfStep',
                                             mic.getMicName(), *args,
                                             prerequisites=prerequisites)
        return micStepId

    def estimateCtfStep(self, micName, *args):
        """ Step function that will be common for all CTF protocols.
        It will take care of re-building the micrograph object from the micDict
        argument and perform any conversion if needed. Then, the function
        _estimateCTF will be called, that should be implemented by each
        CTF estimation protocol.
        """
        mic = self.micDict[micName]
        micDoneFn = self._getMicrographDone(mic)
        micFn = mic.getFileName()

        if self.isContinued() and self._isMicDone(mic):
            self.info("Skipping micrograph: %s, seems to be done" % micFn)
            return

        # Clean old finished files
        pwutils.cleanPath(micDoneFn)
        self.info("Estimating CTF of micrograph: %s " % mic.getObjId())
        self._estimateCTF(mic, *args)

        # Mark this mic as finished
        open(micDoneFn, 'w').close()

    def _estimateCTF(self, mic, *args):
        """ Do the CTF estimation with the specific program
        and the parameters required.
        Params:
         micFn: micrograph filename
         micDir: micrograph directory
        """
        raise Exception(pwutils.Message.ERROR_NO_EST_CTF)

    def reEstimateCtfStep(self, micId):
        """ CTF - re-estimation that is common for all programs.
        The _restimateCTF function will be called with proper parameters.
        """
        ctf = self.recalculateSet[micId]
        mic = self.micDict[micId]
        micDoneFn = self._getMicrographDone(mic)
        # Clean old finished files
        pwutils.cleanPath(micDoneFn)
        self.info("Estimating CTF of micrograph: %s " % mic.getObjId())
        self._reEstimateCTF(mic, ctf)

        # Mark this mic as finished
        open(micDoneFn, 'w').close()

    def _reEstimateCTF(self, mic, ctf):
        """ Do the re-estimation of this mic (original one)
        and the parameters that comes in the comment field of the ctf object.
        Params:
         micFn: micrograph filename
         micDir: micrograph directory
        """
        raise Exception(pwutils.Message.ERROR_NO_EST_CTF)

    # Group of functions to estimate several micrographs if the batch size is
    # defined. In some programs it might be more efficient to estimate many
    # at once and not one by one

    def _insertCtfListStep(self, micList, prerequisites, *args):
        """ Basic method to insert an estimation step for a given micrograph. """
        micNameList = [mic.getMicName() for mic in micList]
        micStepId = self._insertFunctionStep('estimateCtfListStep',
                                             micNameList, *args,
                                             prerequisites=prerequisites)
        return micStepId

    def estimateCtfListStep(self, micNameList, *args):
        micList = []

        for micName in micNameList:
            mic = self.micDict[micName]
            micDoneFn = self._getMicrographDone(mic)
            if self.isContinued() and self._isMicDone(mic):
                self.info("Skipping micrograph: %s, seems to be done"
                          % mic.getFileName())
            else:
                # Clean old finished files
                pwutils.cleanPath(micDoneFn)
                micList.append(mic)

        self.info("Estimating CTF for micrographs: %s"
                  % [mic.getObjId() for mic in micList])
        self._estimateCtfList(micList, *args)

        for mic in micList:
            # Mark this mic as finished
            open(self._getMicrographDone(mic), 'w').close()

    def _estimateCtfList(self, micList, *args):
        """ This function can be implemented by subclasses if it is a more
        efficient way to estimate many micrographs at once.
         Default implementation will just call the _estimateCTF. """
        for mic in micList:
            self._estimateCTF(mic, *args)

    def _createCtfModel(self, mic, updateSampling=False):
        """ This should be implemented in subclasses
        in order to create a CTF model from program results.
        """
        pass

    def createOutputStep(self):
        """ This function is shared by Xmipp and CTFfind
        estimation, or recalculate, protocols.
        if is recalculate, it will iterated for each CTF model, see
        if was recalculated and update with new defocus values.
        Else, the function that should be implemented in each subclass.
        """
        if self.recalculate:
            ctfSet = self._createSetOfCTF("_recalculated")
            prot = self.continueRun.get() or self
            if hasattr(prot, ProtCTFMicOutputs.outputCTF.name):
                micSet = prot.outputCTF.getMicrographs()
                # We suppose this is reading the ctf selection
                # (with enabled/disabled) to only consider the enabled ones
                # in the final SetOfCTF
                # TODO: maybe we can remove the need of the extra text file
                # with the recalculate parameters
                newCount = 0
                for ctfModel in self.recalculateSet:
                    if ctfModel.isEnabled() and ctfModel.getObjComment():
                        mic = ctfModel.getMicrograph()
                        # Update the CTF models that where recalculated and append
                        # later to the set, we don't want to copy the id here since
                        # it is already correct
                        newCtf = self._createCtfModel(mic, updateSampling=False)
                        ctfModel.copy(newCtf, copyId=False)
                        ctfModel.setEnabled(True)
                        newCount += 1
                    ctfSet.append(ctfModel)
                ctfSet.setMicrographs(micSet)
                self._defineOutputs(**{ProtCTFMicOutputs.outputCTF.name:ctfSet})
                self._defineCtfRelation(micSet, ctfSet)
                self._computeDefocusRange(ctfSet)
                self.summaryVar.set("CTF Re-estimation of %d micrographs"
                                    % newCount)
            else:
                raise Exception(
                    pwutils.redStr("The outputCTF do not exist, all CTFs failed."))
        else:
            self._createOutputStep()
            if self.outputCTF.getSize() == 0:
                raise Exception(pwutils.redStr("outputCTF has size zero, all CTFs failed."
                                               "Please review processing steps above."))

    # -------------------------- INFO functions -------------------------------
    def _summary(self):
        summary = []

        if self.recalculate:
            if self.isFinished():
                if self.summaryVar.hasValue():
                    summary.append(self.summaryVar.get())
            else:
                summary.append(pwutils.Message.TEXT_NO_CTF_READY)
        else:
            if not hasattr(self, ProtCTFMicOutputs.outputCTF.name):
                summary.append(pwutils.Message.TEXT_NO_CTF_READY)
            else:
                summary.append("CTF estimation of %d micrographs."
                               % self.inputMicrographs.get().getSize())

        return summary

    def _methods(self):
        methods = []

        if hasattr(self, ProtCTFMicOutputs.outputCTF.name) and self.isFinished():
            methods.append(self.methodsVar.get())
        else:
            methods.append(pwutils.Message.TEXT_NO_CTF_READY)

        return methods

    # -------------------------- UTILS functions ------------------------------
    def _defineCtfParamsDict(self):
        """ This function define a dictionary with parameters used
        for CTF estimation that are common for all micrographs. """
        # Get pointer to input micrographs
        inputMics = self.getInputMicrographs()
        acq = inputMics.getAcquisition()
        sampling = inputMics.getSamplingRate()
        downFactor = self.getAttributeValue('ctfDownFactor', 1.0)
        if downFactor != 1.0:
            sampling *= downFactor
        self._params = {'voltage': acq.getVoltage(),
                        'sphericalAberration': acq.getSphericalAberration(),
                        'magnification': acq.getMagnification(),
                        'ampContrast': acq.getAmplitudeContrast(),
                        'samplingRate': sampling,
                        'scannedPixelSize': inputMics.getScannedPixelSize(),
                        'windowSize': self.windowSize.get(),
                        'lowRes': self.lowRes.get(),
                        'highRes': self.highRes.get(),
                        # Convert from microns to Angstroms
                        'minDefocus': self.minDefocus.get() * 1e+4,
                        'maxDefocus': self.maxDefocus.get() * 1e+4
                        }

    def getCtfParamsDict(self):
        """ Return a copy of the global params dict,
        to avoid overwriting values. """
        return self._params

    def getRecalCtfParamsDict(self, ctfModel):
        """ This function get the acquisition info of the micrographs"""
        mic = ctfModel.getMicrograph()

        acq = mic.getAcquisition()
        mag = acq.getMagnification()
        scannedPixelSize = mic.getSamplingRate() * mag / 10000
        return {'voltage': acq.getVoltage(),
                'sphericalAberration': acq.getSphericalAberration(),
                'magnification': mag,
                'ampContrast': acq.getAmplitudeContrast(),
                'scannedPixelSize': scannedPixelSize,
                'samplingRate': mic.getSamplingRate()
                }

    def _ctfCounter(self, values):
        """ This function return the number of CTFs that was recalculated.
        """
        numberOfCTF = len(values) / 2
        msg = "CTF Re-estimation of %d micrographs" % numberOfCTF
        self.summaryVar.set(msg)

    def _getInputCtf(self):
        if self.continueRecal:
            sqliteFile = self._getPath()
        #             return self.outputCTF.get()
        else:
            return self.inputCtf.get()

    def _iterMicrographs(self, inputMics=None):
        """ Iterate over micrographs and yield
        micrograph name. """
        if inputMics is None:
            inputMics = self.getInputMicrographs()

        for mic in inputMics:
            micFn = mic.getFileName()
            yield micFn, mic

    def _computeDefocusRange(self, ctfSet):
        """ Compute the minimum and maximu defocus in a set of CTFs.
        The protocol methodsVar will be updated with new values.

        Params:
            ctfSet: the set of CTFs to compute min and max
        """
        defocusList = []

        for ctf in ctfSet:
            defocusList.append(ctf.getDefocusU())
            defocusList.append(ctf.getDefocusV())

        minD = min(defocusList) / 10000.
        maxD = max(defocusList) / 10000.

        self.methodsVar.set("Estimated  defocus range defocus was"
                            " %0.3f - %0.3f microns. " % (minD, maxD))

        self._store(self.methodsVar)

    def _defocusMaxMin(self, defocusList):
        """ This function return the minimum and maximum of the defocus
        of a SetOfMicrographs.
        """
        raise Exception("DEPRECATED")

    def getInputMicrographsPointer(self):
        return self.inputMicrographs

    def getInputMicrographs(self):
        return self.getInputMicrographsPointer().get()

    def _getCtfArgs(self):
        """ Should be implemented in sub-classes to define the argument
        list that should be passed to the estimation step function.
        """
        return []

    # ------ Methods for Streaming CTF --------------
    def _stepsCheck(self):
        # To allow streaming ctf estimation we need to detect:
        #   1) new micrographs ready to be estimated
        #   2) new output ctfs that have been produced and add then
        #      to the output set.

        # For now the streaming is not allowed for recalculate CTF
        if self.recalculate:
            return
        self._checkNewInput()
        self._checkNewOutput()

    def _checkNewInput(self):
        # Check if there are new micrographs to process from the input set
        localFile = self.getInputMicrographs().getFileName()
        now = datetime.now()
        self.lastCheck = getattr(self, 'lastCheck', now)
        mTime = datetime.fromtimestamp(getmtime(localFile))
        self.debug('Last check: %s, modification: %s'
                   % (pwutils.prettyTime(self.lastCheck),
                      pwutils.prettyTime(mTime)))
        # If the input micrographs.sqlite have not changed since our last check,
        # it does not make sense to check for new input data
        if self.lastCheck > mTime and hasattr(self, 'listOfMics'):
            return None

        self.lastCheck = now
        # Open input micrographs.sqlite and close it as soon as possible
        micDict, self.streamClosed = self._loadInputList()
        newMics = micDict.values()
        outputStep = self._getFirstJoinStep()

        if newMics:
            fDeps = self._insertNewMicsSteps(newMics)
            if outputStep is not None:
                outputStep.addPrerequisites(*fDeps)
            self.updateSteps()

    def _checkNewOutput(self):
        if getattr(self, 'finished', False):
            return
        # Load previously done items (from text file)
        doneList = self._readDoneList()
        # Check for newly done items
        listOfMics = self.micDict.values()
        nMics = len(listOfMics)
        newDone = [m for m in listOfMics
                   if m.getObjId() not in doneList and self._isMicDone(m)]

        # Update the file with the newly done mics
        # or exit from the function if no new done mics
        self.debug('_checkNewOutput: ')
        self.debug('   listOfMics: %s, doneList: %s, newDone: %s'
                   % (nMics, len(doneList), len(newDone)))

        allDone = len(doneList) + len(newDone)
        # We have finished when there is not more input mics (stream closed)
        # and the number of processed mics is equal to the number of inputs
        self.finished = self.streamClosed and allDone == nMics
        streamMode = pwobj.Set.STREAM_CLOSED if self.finished else pwobj.Set.STREAM_OPEN
        self.debug('   streamMode: %s newDone: %s' % (streamMode,
                                                      not (newDone == [])))

        if newDone:
            newDoneUpdated = self._updateOutputCTFSet(newDone, streamMode)
            self._writeDoneList(newDoneUpdated)
        elif not self.finished:
            # If we are not finished and no new output have been produced
            # it does not make sense to proceed and updated the outputs
            # so we exit from the function here

            # Maybe it would be good idea to take a snap to avoid
            # so much IO if this protocol does not have much to do now
            if allDone == nMics:
                self._streamingSleepOnWait()

            return

        self.debug('   finished: %s ' % self.finished)
        self.debug('        self.streamClosed (%s) AND' % self.streamClosed)
        self.debug('        allDone (%s) == len(self.listOfMics (%s)'
                   % (allDone, nMics))

        if self.finished:  # Unlock createOutputStep if finished all jobs
            self._updateStreamState(streamMode)
            outputStep = self._getFirstJoinStep()
            if outputStep and outputStep.isWaiting():
                outputStep.setStatus(pwcts.STATUS_NEW)

    def _loadInputList(self):
        """ Load the input set of micrographs that are ready to be estimated. """
        return self._loadSet(self.getInputMicrographs(), emobj.SetOfMicrographs,
                             lambda mic: mic.getMicName())

    def _loadSet(self, inputSet, SetClass, getKeyFunc):
        """ Load a given input set if their items are not already present
        in the self.micDict.
        This can be used to load new micrographs for estimation as well as
        new CTF (if used) in streaming.
        """
        setFn = inputSet.getFileName()
        self.debug("Loading input db: %s" % setFn)
        updatedSet = SetClass(filename=setFn)
        updatedSet.loadAllProperties()
        newItemDict = OrderedDict()
        for item in updatedSet:
            micKey = getKeyFunc(item)
            if micKey not in self.micDict:
                newItemDict[micKey] = item.clone()
        streamClosed = updatedSet.isStreamClosed()
        updatedSet.close()
        self.debug("Closed db.")

        return newItemDict, streamClosed

    def _updateOutputCTFSet(self, micList, streamMode):
        doneFailed = []
        micDoneList = [mic for mic in micList]
        # Do no proceed if there is not micrograph ready
        if not micDoneList:
            return []

        outputName = ProtCTFMicOutputs.outputCTF.name
        outputCtf = getattr(self, outputName, None)

        # If there is not outputCTF yet, it means that is the first
        # time we are updating output CTFs, so we need to first create
        # the output set
        firstTime = outputCtf is None

        if firstTime:
            outputCtf = self._createSetOfCTF()
            outputCtf.setMicrographs(self.getInputMicrographsPointer())
        else:
            outputCtf.enableAppend()

        for micFn, mic in self._iterMicrographs(micList):
            try:
                ctf = self._createCtfModel(mic)
                outputCtf.append(ctf)
            except Exception as ex:
                print(pwutils.yellowStr("Missing CTF?: Couldn't update CTF set with mic: %s" % micFn))
                doneFailed.append(mic)

        self.debug(" _updateOutputCTFSet Stream Mode: %s " % streamMode)
        self._updateOutputSet(outputName, outputCtf, streamMode)
        if doneFailed:
            self._writeFailedList(doneFailed)

        if firstTime:  # define relation just once
            # Using a pointer to define the relations is more robust to
            # scheduling and id changes between the protocol run.db and
            # the main project database.get
            self._defineCtfRelation(self.getInputMicrographsPointer(),
                                    outputCtf)

        return micDoneList

    def _updateStreamState(self, streamMode):
        outputName = ProtCTFMicOutputs.outputCTF.name
        outputCtf = getattr(self, outputName, None)

        # If there are not outputCTFs yet, it means that is the first
        # time we are updating output CTF, so we need to first create
        # the output set
        firstTime = outputCtf is None

        if firstTime:
            outputCtf = self._createSetOfCTF()
        else:
            outputCtf.enableAppend()

        self.debug(" _updateStreamState Stream Mode: %s " % streamMode)
        self._updateOutputSet(outputName, outputCtf, streamMode)

    def _readDoneList(self):
        """ Read from a text file the id's of the items that have been done. """
        doneFile = self._getAllDone()
        doneList = []
        # Check what items have been previously done
        if exists(doneFile):
            with open(doneFile) as f:
                doneList += [int(line.strip()) for line in f]
        return doneList

    def _writeDoneList(self, micList):
        """ Write to a text file the items that have been done. """
        doneFile = self._getAllDone()

        if not exists(doneFile):
            pwutils.makeFilePath(doneFile)

        with open(doneFile, 'a') as f:
            for mic in micList:
                f.write('%d\n' % mic.getObjId())

    def _isMicDone(self, mic):
        """ A mic is done if the marker file exists. """
        return exists(self._getMicrographDone(mic))

    def _getAllDone(self):
        return self._getExtraPath('DONE', 'all.TXT')

    def _getAllFailed(self):
        return self._getExtraPath('FAILED_all.TXT')

    def _getMicrographDir(self, mic):
        """ Return an unique dir name for results of the micrograph. """
        return self._getTmpPath('mic_%04d' % mic.getObjId())

    def _getMicrographDone(self, mic):
        """ Return the file that is used as a flag of termination. """
        return self._getExtraPath('DONE', 'mic_%06d.TXT' % mic.getObjId())

    def _writeFailedList(self, micList):
        """ Write to a text file the items that have failed. """
        with open(self._getAllFailed(), 'a') as f:
            for mic in micList:
                f.write('%d\n' % mic.getObjId())

    def _readFailedList(self):
        """ Read from a text file the id's of the items that have failed. """
        failedFile = self._getAllFailed()
        failedList = []
        if exists(failedFile):
            with open(failedFile) as f:
                failedList += [int(line.strip()) for line in f]

        return failedList


class ProtPreprocessMicrographs(ProtMicrographs):
    pass



#!/usr/bin/env python
# ***************************************************************************
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
from PIL import Image

import pyworkflow.tests as pwtests
import pwem.objects as emobj
from pwem.protocols import ProtSetFilter, EMProtocol

OUTPUT_COORDINATES = "outputCoordinates"

class TestSetFilter(pwtests.BaseTest):
    """Run different tests related to the editor set protocol."""
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)


    def createCoorSetProtocol(self):
        MICNUMBER = 2
        MICSIZE = 1000

        # create dummy protocol
        dummyProt = self.newProtocol(EMProtocol)
        dummyProt.setObjLabel('dummy protocol')
        self.launchProtocol(dummyProt)

        # create set of micrographs
        #   create two tif image
        matrix = np.zeros((MICSIZE,MICSIZE), np.uint8)
        R = np.linspace(0, MICSIZE // 2 -2, 20)
        # note we have 0 and 2*pi for each radius
        THETA = np.linspace(0, 2* np.pi, 30)
        radii, thetas = np.meshgrid(R, THETA)
        X = (R * np.cos(thetas)).astype(int) + MICSIZE // 2
        Y = (R * np.sin(thetas)).astype(int) + MICSIZE // 2
        # TODO: next double loop should be equivalen to
        # the more elegant solution
        # pairlist = np.vstack(list(map(np.ravel, (X,Y)))).T
        # matrix[pairList] = 255 but it is not

        for xx, yy in zip(X,Y):
            for x, y in zip(xx, yy):
                matrix[y][x] = 255

        fn = dummyProt._getExtraPath('mic_000001.tif')
        Image.fromarray(matrix).save(fn)
        fn = dummyProt._getExtraPath('mic_000002.tif')
        Image.fromarray(matrix).save(fn)
        micSet = dummyProt._createSetOfMicrographs()
        mic = emobj.Micrograph()

        for i in range(MICNUMBER):
            mic.setFileName\
                (dummyProt._getExtraPath("mic_%06d.tif" % (i%2 + 1 )))
            mic.setMicName("mic_%06d.tif" % (i%2 + 1 ))
            mic.setSamplingRate(1.)
            micSet.append(mic)
            mic.cleanObjId()
        micSet.setSamplingRate(1.)
        micSet.write()

        coordSet = dummyProt._createSetOfCoordinates(micSet)
        coordSet.setBoxSize(10)
        coord = emobj.Coordinate()

        for xx, yy in zip(X,Y):
            for x, y in zip(xx, yy):
                for mic in range(1, MICNUMBER + 1):
                    coord.setX(x)
                    coord.setY(y)
                    coord.setMicId(mic)
                    coord.setMicName("mic_%06d.tif" % (mic))
                    coordSet.append(coord)
                    coord.cleanObjId()
        coordSet.write()
        outputArgs = {OUTPUT_COORDINATES: coordSet,
                      'outputMic': micSet}
        dummyProt._defineOutputs(**outputArgs)
        dummyProt._store()
        return dummyProt

    def testOperation(self):
        """Make a trivial operation. keep coordinates with xcoor > 200"""
        dummyProt = self.createCoorSetProtocol()
        protSetFilter = self.newProtocol(ProtSetFilter,
                                          objLabel="operate")
        protSetFilter.inputSet.set(dummyProt)
        protSetFilter.inputSet.setExtended(OUTPUT_COORDINATES)
        protSetFilter.operation.set(protSetFilter.CHOICE_FORMULA)
        protSetFilter.formula.set('item._x.get()>200')
        self.launchProtocol(protSetFilter)
        for item in protSetFilter.outputCoordinates:
            self.assertGreater(item.getX(), 100)
        self.assertEqual(len(dummyProt.outputCoordinates), 1200)
        self.assertEqual(len(protSetFilter.outputCoordinates), 1100)

    def testCenter(self):
        """Remove coordinates closer to center less than 200"""
        dummyProt = self.createCoorSetProtocol()
        protSetFilter = self.newProtocol(ProtSetFilter,
                                         objLabel="center")
        protSetFilter.inputSet.set(dummyProt)
        protSetFilter.inputSet.setExtended(OUTPUT_COORDINATES)
        protSetFilter.operation.set(protSetFilter.CHOICE_DISTANCE_CENTER)
        protSetFilter.distance.set(200)
        self.launchProtocol(protSetFilter)
        self.assertEqual(len(protSetFilter.outputCoordinates), 720)
        self.assertEqual(len(dummyProt.outputCoordinates), 1200)

    def testDistance(self):
        """Remove coordinates closer to center less than 200"""
        dummyProt = self.createCoorSetProtocol()
        protSetFilter = self.newProtocol(ProtSetFilter,
                                         objLabel="distance- keepfirst true")
        protSetFilter.inputSet.set(dummyProt)
        protSetFilter.inputSet.setExtended(OUTPUT_COORDINATES)
        protSetFilter.operation.set(protSetFilter.CHOICE_DISTANCE_BETWEEN_COORDS)
        protSetFilter.distance.set(10)
        protSetFilter.keepFirst.set(True)
        self.launchProtocol(protSetFilter)
        self.assertEqual(len(protSetFilter.outputCoordinates), 1048)
        self.assertEqual(len(dummyProt.outputCoordinates), 1200)
        protSetFilter = self.newProtocol(ProtSetFilter,
                                         objLabel="distance- keepfirst false")
        protSetFilter.inputSet.set(dummyProt)
        protSetFilter.inputSet.setExtended(OUTPUT_COORDINATES)
        protSetFilter.operation.set(protSetFilter.CHOICE_DISTANCE_BETWEEN_COORDS)
        protSetFilter.distance.set(10)
        protSetFilter.keepFirst.set(False)
        self.launchProtocol(protSetFilter)
        self.assertEqual(len(protSetFilter.outputCoordinates), 1008)
        self.assertEqual(len(dummyProt.outputCoordinates), 1200)

    def testRanking(self):
        """Test ranking filtering"""
        dummyProt = self.createCoorSetProtocol()
        self.addFilterSetProtocol(dummyProt, '10%', dummyProt.outputCoordinates.getSize()*0.1)
        self.addFilterSetProtocol(dummyProt, '-0.3', dummyProt.outputCoordinates.getSize() * 0.3)

        self.addFilterSetProtocol(dummyProt, '10', 10)
        self.addFilterSetProtocol(dummyProt, '-20', 20)

    def addFilterSetProtocol(self, inputProt, threshold, expectedSize):

        protSetFilter = self.newProtocol(ProtSetFilter,
                                          objLabel="ranking %s" % threshold)
        protSetFilter.inputSet.set(inputProt)
        protSetFilter.inputSet.setExtended(OUTPUT_COORDINATES)
        protSetFilter.operation.set(protSetFilter.CHOICE_RANKED)
        protSetFilter.threshold.set(threshold)
        protSetFilter.rankingField.set('_x')
        self.launchProtocol(protSetFilter)
        self.assertSetSize(protSetFilter.outputCoordinates, expectedSize)


# **************************************************************************
# *
# * Authors:     Pablo Conesa(pconesa@cnb.csic.es)
# *              Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.protocol.params as params
from pwem.protocols import EMProtocol
from pwem.objects.data import SetOfCoordinates
import numpy as np

class ProtSetFilter(EMProtocol):
    """
    Protocol to filter sets based on its attributes through an expression that
    should evaluate to true or false. Some predefined expresions are stored (i.e.
    distance to center, distance between coordinates)
    """
    _label = 'filter set'
    CHOICE_FORMULA = 0
    CHOICE_DISTANCE_CENTER = 1
    CHOICE_DISTANCE_BETWEEN_COORDS = 2
    CHOICE_RANKED = 3
    CHOICE_LABEL = {CHOICE_FORMULA: 'formula',
                    CHOICE_DISTANCE_CENTER: 'distance to center',
                    CHOICE_DISTANCE_BETWEEN_COORDS:
                        'distance between particles',
                    CHOICE_RANKED: 'ranking'}

    def _defineParams(self, form):
        """
        Defines the parameters the protocol form will show and its behaviour
        :param form:
        """
        form.addSection(label='Input')
        form.addParam('inputSet', params.PointerParam, pointerClass='EMSet',
                      label='Set to filter',
                      help='Set which items will be filtered.')
        form.addParam('operation', params.EnumParam,
                      choices=[self.CHOICE_LABEL[self.CHOICE_FORMULA],
                               self.CHOICE_LABEL[self.CHOICE_DISTANCE_CENTER],
                               self.CHOICE_LABEL[self.CHOICE_DISTANCE_BETWEEN_COORDS],
                               self.CHOICE_LABEL[self.CHOICE_RANKED]
                               ],
                      default = self.CHOICE_FORMULA,
                      label="Select operation",
                      help="Select operation to be performed in the set.\n"
                           " *distance to center* keep coordinates that are farther from"
                           " the center than a given value.\n"
                           " *distance-between-coordinates* if two coordinates \n"
                           " are closer than a given value then keep only one."
                           " *formula* will apply the formula to the chosen attribute"
                           " (i.e, resolution less than 2 )"
                      )

        form.addParam('formula', params.StringParam, label="Passing formula",
                      condition = "operation==%d" % self.CHOICE_FORMULA,
                      help='A python code compatible with eval that should evaluate to True or False, where item represents each of '
                           'the elements of the set. E.g.: item._resolution.get() < 4).'
                           'You could also use modules like "import numpy;  item._resolution .... "')
        form.addParam('distance', params.FloatParam, default=0,
                      condition="operation==%d or operation==%d" %
                                (self.CHOICE_DISTANCE_CENTER,
                                 self.CHOICE_DISTANCE_BETWEEN_COORDS),
                      label="distance (A)",
                      help="distance from coordinates to center or " \
                           "distance between coordinates ")
        form.addParam('keepFirst', params.BooleanParam, default=True,
                      condition="operation==%d" %
                                self.CHOICE_DISTANCE_BETWEEN_COORDS,
                      label="keep first coordinate",
                      help="If 2 or more coordinates are closer than distance"
                           "keep the first one or delete all"
                      )

        form.addParam('threshold', params.StringParam, label="Threshold: ",
                      condition="operation==%d" % (self.CHOICE_RANKED),
                      help='Number/proportion of items to keep:\n\tNumber: n>=1 \n\tProportion: 0<n<1\n\tPercentage: n%\n\n'
                           'Higher/lower values of the attribute: \n\tHigher: positive number\n\tLower: negative number\n\n'
                           'e.g: "-10%" == "-0.1" == 10% of the items with lower values\n'
                           'e.g: "5" == 5 items with higher values')
        form.addParam('rankingField', params.StringParam, label="Ranking field: ",
                      condition="operation==%d" % self.CHOICE_RANKED,
                      help='Attribute to sort the set by.')

    def _insertAllSteps(self):
        operation = self.operation.get()
        if operation == self.CHOICE_FORMULA:
            self._insertFunctionStep(self.formulaStep)
        elif operation == self.CHOICE_DISTANCE_CENTER:
            self._insertFunctionStep(self.distanceCenterStep)
        elif operation == self.CHOICE_DISTANCE_BETWEEN_COORDS:
            self._insertFunctionStep(self.distanceBetweenCoorStep)
        elif operation == self.CHOICE_RANKED:
            self._insertFunctionStep(self.rankingStep)

    def createOutput(self, modifiedSet):
        # TODO: copyInfo does not copy the set of micrographs
        # associate to the setOfCoordinates
        if hasattr(modifiedSet, "setMicrographs"):
            modifiedSet.setMicrographs(self.inputSet.get().getMicrographs())
        outputArgs = {self.inputSet.getExtended(): modifiedSet}
        self._defineOutputs(**outputArgs)

    def distanceCenterStep(self):
        " compute distance of center to coordinates. Filter if it is two close "
        inputSet = self.inputSet.get()
        modifiedSet = inputSet.create(self._getExtraPath())
        modifiedSet.copyInfo(inputSet)
        mic = inputSet.getMicrographs().getFirstItem()
        micXcenter = mic.getDim()[0] // 2
        micYcenter = mic.getDim()[0] // 2
        sampling = mic.getSamplingRate()
        distance = self.distance.get() / sampling
        distance2 = distance * distance
        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            x = item.getX()
            y = item.getY()
            diffX =  (x-micXcenter)
            diffY =  (y-micYcenter)
            if diffX * diffX + diffY * diffY > distance2:
                modifiedSet.append(item)

        self.createOutput(modifiedSet)

    def distanceBetweenCoorStep(self):
        """ filter by distance between coordinates. If they are too
        close do not keep them"""
        inputSet = self.inputSet.get()
        keepFirstNot = not self.keepFirst.get()
        modifiedSet = inputSet.create(self._getExtraPath())
        modifiedSet.copyInfo(inputSet)
        # set of micrographs is not copied in copyInfo
        mic = inputSet.getMicrographs().getFirstItem()
        sampling = mic.getSamplingRate()
        distance = self.distance.get() / sampling
        distance2 = distance * distance

        # duplicate setOfCoordinates in memory
        coordList = []
        oldMicId=-1
        # If coordenates are sorted by micId then
        # we can process much faster
        useBreak = True
        for item in inputSet.iterItems():
            micId = item.getMicId()
            coordList.append([item.getX(), item.getY(), micId])
            if oldMicId > micId:
                useBreak = False
            oldMicId = micId
        keepList = np.full((len(inputSet)), True, dtype=bool)
        # search for close coordinates
        # this one remove both particles
        for start1, coord1 in enumerate(coordList):
            for start2, coord2 in enumerate(coordList[start1+1:]):
                if (coord1[2] != coord2[2]):  # compare micIds
                    if useBreak:
                        break
                    else:
                        continue
                xd = coord1[0] - coord2[0]
                yd = coord1[1] - coord2[1]
                if (xd*xd + yd*yd) < distance2:
                    if keepFirstNot:
                        keepList[start1] = False
                    keepList[start2+start1+1] = False

        for condition, coord in zip(keepList, inputSet):
            if condition:
                modifiedSet.append(item)

        self.createOutput(modifiedSet)


    def formulaStep(self):
        """
        Goes through all items in the input set and applies the formula to each of them using exec.
        Complex python code could be run separating lines with ;  To use numpy you could do
        import numpy; item._resolution.set(numpy.random.randint(10))
        If result is True, item will be transferred to the output set
        """
        inputSet = self.inputSet.get()
        modifiedSet = inputSet.create(self._getExtraPath())
        modifiedSet.copyInfo(inputSet)

        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            exec("item.setEnabled(%s)"% self.formula.get())
            if item.isEnabled():
                modifiedSet.append(item)
        # TODO: copyInfo does not copy the set of micrographs
        # associate to the setOfCoordinates
        if isinstance(modifiedSet, SetOfCoordinates):
            modifiedSet.setMicrographs(inputSet.getMicrographs())
        self.createOutput(modifiedSet)

    def rankingStep(self):
        """
        Goes through all items in the input set and takes the number/proportion of items with a higher/lower value
        of the chosen attribute
        """
        inputSet = self.inputSet.get()
        finalNumber, direc = self.parseTopRankParam()
        modifiedSet = self.getTopRankItems(self.rankingField.get(), inputSet, finalNumber, direc)
        self.createOutput(modifiedSet)

    def parseTopRankParam(self):
        direc = 'DESC'
        inputSet, threshold = self.inputSet.get(), self.threshold.get().strip()
        # Filed with % at the end
        if threshold.endswith('%'):
            perc = float(threshold[:-1])
            finalNumber = round(perc * len(inputSet) / 100)

        # percentage specified in decimal format.
        elif -1 < float(threshold) < 1:
            prop = float(threshold)
            finalNumber = round(prop * len(inputSet))

        # Else integers (positive or negative).
        else :
            finalNumber = int(float(threshold))

        # If negative
        if finalNumber < 0:
            finalNumber = abs(finalNumber)
            direc = 'ASC'

        return finalNumber, direc

    def getTopRankItems(self, attribute, iSet, finalNumber, direc='ASC'):
        modifiedSet = iSet.createCopy(self._getExtraPath(), copyInfo=True)
        for item in iSet.iterItems(orderBy=attribute, direction=direc, limit=finalNumber):
            modifiedSet.append(item.clone())
        return modifiedSet


    def _validate(self):
        errors = []
        inputSet = self.inputSet.get()
        operation = self.operation.get()
        if operation == self.CHOICE_DISTANCE_CENTER or operation == self.CHOICE_DISTANCE_BETWEEN_COORDS:
            if not isinstance(inputSet, SetOfCoordinates):
                errors.append("The input data set is not a set of coordinates")
        elif operation == self.CHOICE_RANKED:
            param = self.threshold.get().strip()
            perc = False
            try:
                if param.endswith('%'):
                    param = param[:-1]
                    perc=True
                param = float(param)
                if perc:
                    if abs(param) > 100:
                        errors.append("Percentage cannot be higher than 100")
            except:
                errors.append("The filter value must be a number or a percentage")
        return errors



# **************************************************************************
# *
# * Authors: Yunior C. Fonseca Reyna    (cfonseca@cnb.csic.es)
# *
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import re
import time

from pyworkflow.protocol.params import  StringParam

from pwem.protocols import EMProtocol


class ProtManualCheckpoint(EMProtocol):
    """
    This protocol is kept running for a time determined by a parameter or until
    the user determines it is convenient. This protocol is useful if we want a
    given protocol to be launched at the time the user sees appropriate.
    """
    _label = 'manual check point'

    def _defineParams(self, form):
        form.addSection(label='Criteria')
        form.addParam('timeout', StringParam, default="1h",
                      label='Time to wait:',
                      help='Time in seconds that the protocol will remain '
                           'running. A correct format is an integer number in '
                           'seconds or the following syntax: {days}d {hours}h '
                           '{minutes}m {seconds}s separated by spaces '
                           'e.g: 1d 2h 20m 15s,  10m 3s, 1h, 20s or 25.')

    def _insertAllSteps(self):
        self._insertFunctionStep(self.waitingStep)

    def waitingStep(self):
        lastTimeCheck = 0
        timeout = self.getTimeOutInSeconds(self.timeout.get())
        finished = False

        while not finished:
            if lastTimeCheck % 300 == 0:
                self.info("Remaining time: %s seconds. Next check in 5 minutes" % str(timeout - lastTimeCheck))
            time.sleep(10)
            lastTimeCheck += 10
            finished = (lastTimeCheck > timeout or
                        self.waitingHasFinished())

    def getTimeOutInSeconds(self, timeOut):
        timeOutFormatRegexList = {'\d+s':1, '\d+m':60, '\d+h':3600,
                                  '\d+d':72000}
        try:
            return int(timeOut)
        except Exception:
            seconds = 0
        for regex, secondsUnit in timeOutFormatRegexList.items():
            matchingTimes = re.findall(regex, timeOut)
            for matchTime in matchingTimes:
                seconds += int(matchTime[:-1]) * secondsUnit

        return seconds

    def _validate(self):
        message = []
        timeInSec = self.getTimeOutInSeconds(self.timeout.get())
        if timeInSec == 0:
            message.append('Time format is wrong. A correct format is an '
                           'integer number in seconds or the following syntax: '
                           '{days}d {hours}h {minutes}m {seconds}s separated by'
                           ' spaces e.g: 1d 2h 20m 15s,  10m 3s, 1h, 20s or 25.')
        return message

    def _getStopWaitingFilename(self):
        return self._getExtraPath("APPROVE_CHECKPOINT.TXT")

    def getActions(self):
        if self.isRunning():
            return [('APPROVE CHECKPOINT', self.stopWait)]
        else:
            return []

    def stopWait(self):
        f = open(self._getStopWaitingFilename(), 'w')
        f.close()

    def waitingHasFinished(self):
        return os.path.exists(self._getStopWaitingFilename())



# ***************************************************************************
# *
# * Authors:     David Maluenda (dmaluenda@cnb.csic.es)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# ***************************************************************************/

import pyworkflow.tests as pwtests

from pwem import Domain
import pwem.protocols as emprot


class TestCtfConsensus(pwtests.BaseTest):
    """ Check if the Xmipp-CTFconsensus rejects CTFs (and the corresponding mics)
        when two CTF estimations give different results,
        and accept when the two estimations give similar results.
    """

    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dataset = pwtests.DataSet.getDataSet('xmipp_tutorial')
        cls.micsFn = cls.dataset.getFile('allMics')

    def checkCTFs(self, protConsensus, refCTFs, refMics, label='',
                  avgCTF=None, MDmerging=False):
        outputCTF = getattr(protConsensus, "outputCTF" + label, None)
        outputMicrographs = getattr(protConsensus, "outputMicrographs" + label, None)

        self.assertIsNotNone(outputCTF,
                             "There was a problem with the CTF-Consensus. "
                             "No outputCTF%s is created." % label)
        self.assertIsNotNone(outputMicrographs,
                             "There was a problem with the CTF-Consensus. "
                             "No outputMicrographs%s is created." % label)

        self.assertEqual(outputCTF.getSize(), refCTFs.getSize(),
                         "The outputCTF%s size is wrong." % label)
        self.assertEqual(outputMicrographs.getSize(), refCTFs.getSize(),
                         "The outputMicrographs%s size is wrong." % label)
        self.assertTupleEqual(outputMicrographs.getDim(), refMics.getDim(),
                              "The outputMicrographs%s dimension is wrong." % label)

        # we will check that the first CTF in the set make sense.
        firstCTF = outputCTF.getFirstItem()

        if not avgCTF:
            refCTF = refCTFs.getFirstItem()
            self.assertTrue(firstCTF.equalAttributes(refCTF),
                            "The outputCTF%s has different attributes "
                            "than the input." % label)
        else:
            self.assertAlmostEqual(avgCTF['defocusU'], firstCTF.getDefocusU(), delta=200,
                                   msg="DefocusU doesn't match when defocus averaging.")
            self.assertAlmostEqual(avgCTF['defocusV'], firstCTF.getDefocusV(), delta=100,
                                   msg="DefocusV doesn't match when defocus averaging.")
            self.assertAlmostEqual(avgCTF['defocusAngle'], firstCTF.getDefocusAngle(), delta=1,
                                   msg="DefocusAngle doesn't match when defocus averaging.")

        if MDmerging:
            for label in ['CritCorr13', 'CritIceness', 'CritCtfMargin',
                          'CritPsdCorr90', 'Q0', 'CritNonAstigmaticValidty',
                          'CritfirstZeroRatio', 'CritFirstZero']:
                MDlabel = '_xmipp_ctf' + label
                self.assertTrue(MDlabel in firstCTF.getObjDict().keys(),
                                "'%s' metadata not found in the result. "
                                "Bad merging" % MDlabel)

    def test1(self):
        # Import a set of micrographs
        protImport = self.newProtocol(emprot.ProtImportMicrographs,
                                      filesPath=self.micsFn,
                                      samplingRate=1.237, voltage=300)
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographs,
                             "There was a problem with the import")

        # Create pure noise micrographs (to force a Discarded consensus set)
        protStream = self.newProtocol(emprot.ProtCreateStreamData,
                                      xDim=9216,
                                      yDim=9441,
                                      nDim=3,
                                      samplingRate=1.237,
                                      setof=2,  # 2 -> SetOfRandomMicrographs
                                      creationInterval=1)
        self.proj.launchProtocol(protStream, wait=False)

        # Computes the CTF with Xmipp
        XmippProtCTFMicrographs = Domain.importFromPlugin('xmipp3.protocols',
                                                          'XmippProtCTFMicrographs',
                                                          doRaise=True)
        protCTF1 = self.newProtocol(XmippProtCTFMicrographs)
        protCTF1.inputMicrographs.set(protImport.outputMicrographs)
        self.proj.launchProtocol(protCTF1, wait=False)

        # Computes the CTF with CTFFind4
        ProtCTFFind = Domain.importFromPlugin('cistem.protocols',
                                              'CistemProtCTFFind', doRaise=True)
        protCTF2 = self.newProtocol(ProtCTFFind)
        protCTF2.inputMicrographs.set(protImport.outputMicrographs)
        self.proj.launchProtocol(protCTF2, wait=False)

        self._waitOutput(protStream, "outputMicrographs")
        # Computes the CTF with CTFFind4 for the noise mics
        protCTF3 = self.newProtocol(ProtCTFFind)
        protCTF3.inputMicrographs.set(protStream.outputMicrographs)
        self.proj.launchProtocol(protCTF3, wait=False)

        # Computes the Consensus of GOOD CTFs
        self._waitOutput(protCTF1, "outputCTF")
        self._waitOutput(protCTF2, "outputCTF")
        XmippProtCTFConsensus = Domain.importFromPlugin('xmipp3.protocols',
                                                        'XmippProtCTFConsensus',
                                                        doRaise=True)
        protCTFcons = self.newProtocol(XmippProtCTFConsensus,
                                       objLabel='default (pass)',
                                       useDefocus=False,
                                       useAstigmatism=False,
                                       useResolution=False,
                                       calculateConsensus=True,
                                       averageDefocus=False,
                                       includeSecondary=False)
        protCTFcons.inputCTF.set(protCTF1.outputCTF)
        protCTFcons.inputCTF2.set(protCTF2.outputCTF)
        self.launchProtocol(protCTFcons)

        protCTF1.outputCTF.load()  # Needed to update the setSize
        self.checkCTFs(protCTFcons,
                       refMics=protImport.outputMicrographs,
                       refCTFs=protCTF1.outputCTF)

        # Computes the Consensus comparing a good CTF to a RANDOM one
        self._waitOutput(protCTF3, "outputCTF")
        protCTFcons2 = self.newProtocol(XmippProtCTFConsensus,
                                        objLabel='default (block)',
                                        useDefocus=False,
                                        useAstigmatism=False,
                                        useResolution=False,
                                        calculateConsensus=True,
                                        averageDefocus=False,
                                        includeSecondary=False)

        protCTFcons2.inputCTF.set(protCTF1.outputCTF)
        protCTFcons2.inputCTF2.set(protCTF3.outputCTF)
        self.launchProtocol(protCTFcons2)
        self.checkCTFs(protCTFcons2,
                       refMics=protImport.outputMicrographs,
                       refCTFs=protCTF1.outputCTF,
                       label="Discarded")

        # Averaging CTF parameters
        protCTFcons3 = self.newProtocol(XmippProtCTFConsensus,
                                        objLabel='defocus average',
                                        useDefocus=False,
                                        useAstigmatism=False,
                                        useResolution=False,
                                        calculateConsensus=True,
                                        averageDefocus=True)
        protCTFcons3.inputCTF.set(protCTF1.outputCTF)
        protCTFcons3.inputCTF2.set(protCTF2.outputCTF)
        self.launchProtocol(protCTFcons3)

        protCTF1.outputCTF.load()  # Needed to update the set
        protCTF2.outputCTF.load()  # Needed to update the set
        ctfAveraged = {'defocusU': 24025.6729,
                       'defocusV': 23610.2071,
                       'defocusAngle': 57.1943}
        self.checkCTFs(protCTFcons3,
                       refMics=protImport.outputMicrographs,
                       refCTFs=protCTF1.outputCTF,
                       avgCTF=ctfAveraged)

        # merging Metadata columns
        protCTFcons4 = self.newProtocol(XmippProtCTFConsensus,
                                        objLabel='metadata merge',
                                        useDefocus=False,
                                        useAstigmatism=False,
                                        useResolution=False,
                                        calculateConsensus=True,
                                        includeSecondary=True)
        protCTFcons4.inputCTF.set(protCTF2.outputCTF)
        protCTFcons4.inputCTF2.set(protCTF1.outputCTF)
        self.launchProtocol(protCTFcons4)

        protCTF1.outputCTF.load()  # Needed to update the set
        protCTF2.outputCTF.load()  # Needed to update the set
        self.checkCTFs(protCTFcons4,
                       refMics=protImport.outputMicrographs,
                       refCTFs=protCTF2.outputCTF,
                       MDmerging=True)
